
#include <unistd.h>
#include <ctype.h>
#include <string.h>

#include "globdef.h"
#include "uidef.h"
#include "fft1def.h"
#include "fft2def.h"
#include "fft3def.h"
#include "sigdef.h"
#include "screendef.h"
#include "seldef.h"
#include "thrdef.h"
#include "txdef.h"
#include "vernr.h"
#include "hwaredef.h"
#include "sdrdef.h"
#include "keyboard_def.h"

#define MAX_SSBPROC_FILES 15
#define MAX_CWPROC_FILES 10
#define SSB_MINFQ_LOW 50
#define SSB_MINFQ_HIGH 3000
#define SSB_MAXFQ_LOW 1000
#define SSB_MAXFQ_HIGH 8000
#define SSB_MINSLOPE -50
#define SSB_MAXSLOPE 50
#define SSB_MINBASS -50
#define SSB_MAXBASS 50
#define SSB_MINTREBLE -50
#define SSB_MAXTREBLE 50
#define SSB_MAX_MICF 120
#define SSB_MAX_MICT 120
#define SSB_MAX_MICGAIN 120
#define SSB_MAX_MICAGC_TIME 1000
#define SSB_MAX_MICOUT_GAIN 80
#define SSB_MIN_RF1_GAIN -50
#define SSB_MAX_RF1_GAIN 120


#define MAX_DYNRANGE 0.0000001 // -70dB for weakest signal

#define TX_DA_MARGIN 0.97
#define TX_WRBLOCK_MAXCNT 100
#define TX_AGC_THRESHOLD 1.

#define TX_SETUP_AD 0
#define TX_SETUP_MUTE 1
#define TX_SETUP_MICAGC 2
#define TX_SETUP_RF1AGC 3
#define TX_SETUP_RF1CLIP 4

#define TX_BAR_AD_LINE 9
#define TX_BAR_MUTE_LINE 11
#define TX_BAR_MICAGC_LINE 13
#define TX_BAR_RF1AGC_LINE 15
#define TX_BAR_RF1CLIP_LINE 17
#define TX_BAR_ALC_LINE 19
#define TXPAR_INPUT_LINE 21

#define MIC_KEY_LINE 12
#define KEY_INDICATOR_LINE (MIC_KEY_LINE+10)
#define TXCW_INPUT_LINE (KEY_INDICATOR_LINE+4)
#define CW_MAX_TIME_CONSTANT 100

#define TX_BAR_SIZE 256

#define MAX_DISPLAY_TYPES 3
#define DISPLAY_HOLD_SIZE 32
#define TG_HSIZ (FG_HSIZ+4*text_width)
#define TG_VSIZ (FG_VSIZ+text_height)


#define MAX_SCREENCOUNT 4

int mic_key_pa;
int mic_key_px;
int screen_count;
int tx_lowest_bin;
int tx_highest_bin;
char *clipfft_mute;
char *cliptimf_mute;
char *alctimf_mute;
float *tx_mic_key;
float micfft_winfac;
int tg_old_y1;
int tg_old_y2;
int tg_old_x1;
int tg_old_x2;
float tx_ssbproc_fqshift;
int mic_key_bufsiz;
int mic_key_block;
int mic_key_mask;

void make_tx_graph(int clear_old);
void init_txmem_cwproc(void);
void make_txproc_filename(void);
void tx_ssb_step8(void);
float tx_ssb_step7(float *prat);
float tx_ssb_step6(float *prat);
void tx_ssb_step5(void);
void tx_ssb_step4(float ampinv,float *t1, float *prat);
void tx_ssb_step3(float *totamp);
int tx_ssb_step2(float *totpwr);


char ssbproc_filename[14]="par_ssbproc00";
char cwproc_filename[13]="par_cwproc00";
char *txproc_filename;

void tx_output(void)
{
char color;
char s[80];
int tone_key, old_tone_key, old_hand_key;
int i, k;
float t1, t2, t3, r2, prat;
tx_setup_flag=FALSE;
old_tone_key=-1;
old_hand_key=-1;
thread_status_flag[THREAD_TX_OUTPUT]=THRFLAG_ACTIVE;
switch (rx_mode)
  {
  case MODE_SSB:
  while(!kill_all_flag && 
                    thread_command_flag[THREAD_TX_OUTPUT] == THRFLAG_ACTIVE)
    {
    lir_sem_wait(SEM_TXINPUT);
    tx_agc_factor=tx_agc_decay*tx_agc_factor+(1-tx_agc_decay);
    micfft_bin_minpower=micfft_bin_minpower_ref*tx_agc_factor*tx_agc_factor;
    micfft_minpower=micfft_minpower_ref*tx_agc_factor*tx_agc_factor;
    k=tx_ssb_step2(&t1);
    if(k!=0)
      {
      tx_ssb_step3(&t1);
      }
// In case AGC has reduced the gain by more than 20 dB, set the
// AGC factor to -20 dB immediately because voice should never
// be as loud. This is to avoid the agc release time constant
// for impulse sounds caused by hitting the microphone etc.       
    if(tx_agc_factor < 0.1)tx_agc_factor=0.1;
    if(k!=0)
      {
      fftback(mic_fftsize, mic_fftn, &micfft[micfft_px], 
                                              micfft_table, micfft_permute);
      r2=0;
      for(i=0; i<mic_fftsize; i++)
        {
        t2=micfft[micfft_px+2*i  ]*micfft[micfft_px+2*i  ]+
           micfft[micfft_px+2*i+1]*micfft[micfft_px+2*i+1];
        if(t2>r2)r2=t2;
        }
      if(r2 > 0.01*MAX_DYNRANGE)
        {
        r2=1/sqrt(r2);
        }
      else
        {
        r2=0;
        }  
      }
    else
      {
      r2=0;
      }  
    tx_ssb_step4(r2,&t1,&prat);
    tx_ssb_step5();
    while(clipfft_px != clipfft_pa)
      {
      tx_ssb_step6(&prat);
      }
    while( ((alctimf_pa-alctimf_pb+alc_bufsiz)&alc_mask) > 3*alc_fftsize)
      {
      tx_ssb_step7(&prat);
      }
    tx_ssb_step8();
    }
  break;

  case MODE_WCW:
  case MODE_NCW:
  case MODE_HSMS:
  case MODE_QRSS:
  while(!kill_all_flag && 
                    thread_command_flag[THREAD_TX_OUTPUT] == THRFLAG_ACTIVE)
    {
    lir_sem_wait(SEM_TXINPUT);

  t1=0;
  t2=0;
  t3=BIG;
  i=0;
  while(mic_key_px != mic_key_pa)
    {
    i++;
    t1+=tx_mic_key[mic_key_px];
    if(t2 < tx_mic_key[mic_key_px])t2=tx_mic_key[mic_key_px];
    if(t3 > tx_mic_key[mic_key_px])t3=tx_mic_key[mic_key_px];
    mic_key_px=(mic_key_px+1)&mic_key_mask;
    }
  t1/=i;  
  sprintf(s,"%.3f",t1);
  lir_text(1,MIC_KEY_LINE+2,s);
  sprintf(s,"%.3f",t2);
  lir_text(12,MIC_KEY_LINE+2,s);
  sprintf(s,"%.3f",t3);
  lir_text(23,MIC_KEY_LINE+2,s);
  if(t1 < 1)
    {
    tone_key=0;
    }
  else
    {
    tone_key=1;
    }  
//#define KEY_INDICATOR_LINE 22
  if(tone_key != old_tone_key)
    {
    old_tone_key=tone_key;
    if(tone_key == 0)
      {
      color=7;
      }
    else
      {
      color=12;
      }
    lir_fillbox(5*text_width, KEY_INDICATOR_LINE*text_height, 10*text_width,
                                                        3*text_height,color);
    lir_text(8,KEY_INDICATOR_LINE+1,"TONE");
    }
  hware_hand_key();
  if(hand_key != old_hand_key)
    {
    old_hand_key=hand_key;
    if(hand_key == 0)
      {
      color=7;
      }
    else
      {
      color=12;
      }
    lir_fillbox(20*text_width, KEY_INDICATOR_LINE*text_height, 10*text_width,
                                                        3*text_height,color);
    lir_text(22,KEY_INDICATOR_LINE+1,"HAND");
    }

    }
  break;
  }

thread_status_flag[THREAD_TX_OUTPUT]=THRFLAG_RETURNED;
while(thread_command_flag[THREAD_TX_OUTPUT] != THRFLAG_NOT_ACTIVE)
  {
  lir_sleep(1000);
  }
}

void cwproc_setup(void)
{
char color;
char *onoff[]={"Disabled", "Enabled"};
char s[256];
int tone_key, old_tone_key, old_hand_key;
int i, k, default_spproc_no;
float t1, t2, t3;
rx_mode=MODE_NCW;
default_spproc_no=tg.spproc_no;
tx_setup_flag=TRUE;
open_tx_output();
open_tx_input();
if(kill_all_flag)return;
init_txmem_cwproc();
if(kill_all_flag)return;
lir_mutex_init();
lir_sleep(50000);
linrad_thread_create(THREAD_TX_INPUT);
restart:;
old_tone_key=-1;
old_hand_key=-1;
clear_screen();
lir_text(20,0,"Setup for CW keying.");
lir_text(0,2,"There are three ways to produce Morse code in Linrad:");
if(txcw.enable_tone_key == 0)settextcolor(7); else settextcolor(14);
sprintf(s,"Tone into the microphone input. (Fast keying)  [%s]",
                    onoff[txcw.enable_tone_key]);
lir_text(0,4,s);
if(txcw.enable_hand_key == 0)settextcolor(7); else settextcolor(14);
sprintf(s,"Parallel port pin 13. (Hand keying)  [%s]",
                                          onoff[txcw.enable_hand_key]);
lir_text(0,5,s);
if(txcw.enable_ascii == 0)settextcolor(7); else settextcolor(14);
sprintf(s,"Computer generated from ascii on keyboard.  [%s]",
                                          onoff[txcw.enable_ascii]);
lir_text(0,6,s);
lir_text(0,8,"Press 'T', 'P' or 'C' to toggle Enable/Disable");
sprintf(s,"Rise time = %.2f ms. ('A' to change)",txcw.rise_time*0.001);
lir_text(0,7,s);
make_txproc_filename();
sprintf(s,"Press 'S' to save %s, 'R' to read.  '+' or '-' to change file no.",
                                                 txproc_filename);
lir_text(0,8,s);

k=screen_last_col;
if(k>255)k=255;
for(i=0; i<k; i++)s[i]='-';
s[k]=0;
lir_text(0,MIC_KEY_LINE,s);
lir_text(0,MIC_KEY_LINE+8,s);
//                         0123456789012345678901234567890
lir_text(0,MIC_KEY_LINE+1,"Average       Max        Min");
lir_text(0,MIC_KEY_LINE+3,
"Set tone level and frequency for numbers well above 1.0 with key down.");
lir_text(0,MIC_KEY_LINE+4,
"Make sure the numbers are well below 1.0 with key up.");
lir_text(0,MIC_KEY_LINE+5,
"Typically Min(key down) > 10.0 and Max(key up) < 0.1.");
sprintf(s,
  "Your Nyquist frequency is %.1f kHz so you should use a keying tone",
                                                     0.0005*ui.tx_ad_speed);
lir_text(0,MIC_KEY_LINE+6,s);
lir_text(0,MIC_KEY_LINE+7,"that is a little lower than this.");



while(!kill_all_flag)
  {
  lir_sem_wait(SEM_TXINPUT);
  t1=0;
  t2=0;
  t3=BIG;
  i=0;
  while(mic_key_px != mic_key_pa)
    {
    i++;
    t1+=tx_mic_key[mic_key_px];
    if(t2 < tx_mic_key[mic_key_px])t2=tx_mic_key[mic_key_px];
    if(t3 > tx_mic_key[mic_key_px])t3=tx_mic_key[mic_key_px];
    mic_key_px=(mic_key_px+1)&mic_key_mask;
    }
  t1/=i;  
  sprintf(s,"%.3f",t1);
  lir_text(1,MIC_KEY_LINE+2,s);
  sprintf(s,"%.3f",t2);
  lir_text(12,MIC_KEY_LINE+2,s);
  sprintf(s,"%.3f",t3);
  lir_text(23,MIC_KEY_LINE+2,s);
  if(t1 < 1)
    {
    tone_key=0;
    }
  else
    {
    tone_key=1;
    }  
//#define KEY_INDICATOR_LINE 22
  if(tone_key != old_tone_key)
    {
    old_tone_key=tone_key;
    if(tone_key == 0)
      {
      color=7;
      }
    else
      {
      color=12;
      }
    lir_fillbox(5*text_width, KEY_INDICATOR_LINE*text_height, 10*text_width,
                                                        3*text_height,color);
    lir_text(8,KEY_INDICATOR_LINE+1,"TONE");
    }
  hware_hand_key();
  if(hand_key != old_hand_key)
    {
    old_hand_key=hand_key;
    if(hand_key == 0)
      {
      color=7;
      }
    else
      {
      color=12;
      }
    lir_fillbox(20*text_width, KEY_INDICATOR_LINE*text_height, 10*text_width,
                                                        3*text_height,color);
    lir_text(22,KEY_INDICATOR_LINE+1,"HAND");
    }
  test_keyboard();
  if(lir_inkey != 0)
    {
    process_current_lir_inkey();
    if(lir_inkey=='X')goto end_tx_setup;
    switch (lir_inkey)
      {
      case 'A':
      lir_text(0,TXCW_INPUT_LINE,"Rise time (ms):");
      t1=lir_get_float(15,TXCW_INPUT_LINE,6,0,CW_MAX_TIME_CONSTANT);
      txcw.rise_time=1000*t1;  
      break;

      case 'T':
      txcw.enable_tone_key^=1;
      txcw.enable_tone_key&=1;
      break;
      
      case 'H':
      txcw.enable_hand_key^=1;
      txcw.enable_hand_key&=1;
      break;

      case 'C':
      txcw.enable_ascii^=1;
      txcw.enable_ascii&=1;
      break;

      case 'S':
      save_tx_parms(TRUE);
      break;

      case 'R':
      clear_screen();
      read_txpar_file();
      break;

      case '+':
      tg.spproc_no++;
      if(tg.spproc_no > MAX_CWPROC_FILES)tg.spproc_no=MAX_CWPROC_FILES;
      break;

      case '-':
      tg.spproc_no--;
      if(tg.spproc_no < 0)tg.spproc_no=0;
      break;
      }
    goto restart;
    }
  }
end_tx_setup:;
close_tx_output();
linrad_thread_stop_and_join(THREAD_TX_INPUT);
close_tx_input();
free(txmem_handle);
lir_mutex_destroy();
tg.spproc_no=default_spproc_no;
}



void make_tx_modes(void)
{
if(ui.tx_ad_bytes == 2)
  {
  if(ui.tx_ad_channels==1)
    {
    tx_input_mode=0;
    tx_input_fragsize=2;
    }
  else
    {
    tx_input_mode=1;
    tx_input_fragsize=4;
    }  
  }
else
  {
  if(ui.tx_ad_channels==1)
    {
    tx_input_mode=2;
    tx_input_fragsize=4;
    }
  else
    {
    tx_input_mode=3;
    tx_input_fragsize=8;
    }  
  }
if(ui.tx_da_bytes == 2)
  {
  if(ui.tx_da_channels==1)
    {
    tx_output_mode=0;
    tx_output_fragsize=2;
    tx_write_bytes=txout_sizhalf*sizeof(short int);
    }
  else
    {
    tx_output_mode=1;
    tx_output_fragsize=4;
    tx_write_bytes=txout_fftsize*sizeof(short int);
    }  
  }
else
  {
  if(ui.tx_da_channels==1)
    {
    tx_output_mode=2;
    tx_output_fragsize=4;
    tx_write_bytes=txout_sizhalf*sizeof(int);
    }
  else
    {
    tx_output_mode=3;
    tx_output_fragsize=8;
    tx_write_bytes=txout_fftsize*sizeof(int);
    }  
  }
}

void init_txmem_cwproc(void)
{
int i;
micsize=ui.tx_ad_speed*.01;
make_power_of_two(&micsize);
micn=0;
i=micsize;
while(i>1)
  {
  i>>=1;
  micn++;
  }
mic_sizhalf=micsize/2;
mictimf_block=ui.tx_ad_channels*micsize;
// Allow the input buffer to hold 80 milliseconds of data (or more)
mictimf_bufsiz=16*mictimf_block;
mictimf_mask=mictimf_bufsiz-1;
tx_read_bytes=ui.tx_ad_bytes*mictimf_block;
// Keying with a tone into the microphone input should
// be done with a tone frequency near, but below the Nyquist frequency
// which is ui.tx_ad_speed/2.
// We do not want faster rise times than 1 ms for a dot length
// of 2 ms minimum or a dot rate of 250 Hz maximum.
// With at least 4 points on the rise time we need a sampling
// rate of 4kHz.
tx_pre_resample=1;
i=1.05*ui.tx_ad_speed/2;
mic_key_block=mictimf_block;
while(i > 4000)
  {
  i/=2;
  mic_key_block/=2;
  tx_pre_resample*=2;
  }
tx_resample_ratio=(float)(2*ui.tx_da_speed)/ui.tx_ad_speed;
mic_key_bufsiz=16*mic_key_block;
mic_key_mask=mic_key_bufsiz-1;
init_memalloc(txmem, MAX_TXMEM_ARRAYS);
if(ui.tx_ad_bytes == 2)
  {
  mem(51,&mictimf_shi,mictimf_bufsiz*sizeof(short int),0);
  }
else
  {
  mem(51,&mictimf_int,mictimf_bufsiz*sizeof(int),0);
  }
mem(52,&tx_mic_key,mic_key_bufsiz*sizeof(float),0);
txmem_size=memalloc((int*)(&txmem_handle),"txmem");
if(txmem_size == 0)
  {
  lirerr(1261);
  return;
  }
make_tx_modes();  
mic_key_pa=0;
mic_key_px=0;
mictimf_pa=0;
mictimf_px=0;
}

void tx_input(void)
{
int i, ia, ib, pa, pb, pc;
int k;
int twosiz, ssb_bufmin;
int nread;
float *z;
float t1;
twosiz=2*micsize;
ssb_bufmin=twosiz*ui.tx_ad_channels;
pb=mictimf_px;
thread_status_flag[THREAD_TX_INPUT]=THRFLAG_ACTIVE;
while(!kill_all_flag && 
                    thread_command_flag[THREAD_TX_INPUT] == THRFLAG_ACTIVE)
  {
  if(ui.tx_ad_bytes == 2)
    {
    nread=read(tx_audio_in,&mictimf_shi[mictimf_pa],tx_read_bytes); 
    }
  else
    {
    nread=read(tx_audio_in,&mictimf_int[mictimf_pa],tx_read_bytes); 
    }
  if(nread != tx_read_bytes)
    {
    lirerr(1283);
    goto tx_input_x;
    }  
  mictimf_pa=(mictimf_pa+mictimf_block)&mictimf_mask;
  switch (rx_mode)
    {
    case MODE_SSB:
    if( ((mictimf_pa-mictimf_px+mictimf_bufsiz)&mictimf_mask) >= ssb_bufmin)
      {
// Copy the input to micfft while multiplying with the window function.
      ib=twosiz-1;
      pa=mictimf_px;
      switch(tx_input_mode)
        {
        case 0:
        pb=(pa+ib)&mictimf_mask;
        for( ia=0; ia<micsize; ia++)
          {
          mic_tmp[mic_permute[ia]]=(float)(mictimf_shi[pa])*mic_win[ia];
          if(abs(mictimf_shi[pa]) > tx_ad_maxamp)
            {
            tx_ad_maxamp=abs(mictimf_shi[pa]);
            }
          mic_tmp[mic_permute[ib]]=(float)(mictimf_shi[pb])*mic_win[ia];
          if(abs(mictimf_shi[pb]) > tx_ad_maxamp)
            {
            tx_ad_maxamp=abs(mictimf_shi[pb]);
            }
          ib--; 
          pa=(pa+1)&mictimf_mask;
          pb=(pb+mictimf_mask)&mictimf_mask;
          }
        break;
        
        case 1:
// In case there are two channels for the microphone, just use one of them!!
        pb=(pa+2*ib)&mictimf_mask;
        for( ia=0; ia<micsize; ia++)
          {
          mic_tmp[mic_permute[ia]]=mictimf_shi[pa]*mic_win[ia];
          if(abs(mictimf_shi[pa]) > tx_ad_maxamp)
            {
            tx_ad_maxamp=abs(mictimf_shi[pa]);
            }
          mic_tmp[mic_permute[ib]]=mictimf_shi[pb]*mic_win[ia];
          if(abs(mictimf_shi[pb]) > tx_ad_maxamp)
            {
            tx_ad_maxamp=abs(mictimf_shi[pb]);
            }
          ib--; 
          pa=(pa+2)&mictimf_mask;
          pb=(pb+mictimf_mask-1)&mictimf_mask;
          }
        break;
        
        case 2:
        pb=(pa+ib)&mictimf_mask;
        for( ia=0; ia<micsize; ia++)
          {
          mic_tmp[mic_permute[ia]]=mictimf_int[pa]*mic_win[ia];
          if(abs(mictimf_int[pa]) > tx_ad_maxamp)
            {
            tx_ad_maxamp=abs(mictimf_int[pa]);
            }
          mic_tmp[mic_permute[ib]]=mictimf_int[pb]*mic_win[ia];
          if(abs(mictimf_int[pb]) > tx_ad_maxamp)
            {
            tx_ad_maxamp=abs(mictimf_int[pb]);
            }
          ib--; 
          pa=(pa+1)&mictimf_mask;
          pb=(pb+mictimf_mask)&mictimf_mask;
          }
        break;
        
        case 3:
// In case there are two channels for the microphone, just use one of them!!
        pb=(pa+2*ib)&mictimf_mask;
        for( ia=0; ia<micsize; ia++)
          {
          mic_tmp[mic_permute[ia]]=mictimf_int[pa]*mic_win[ia];
          if(abs(mictimf_int[pa]) > tx_ad_maxamp)
            {
            tx_ad_maxamp=abs(mictimf_int[pa]);
            }
          mic_tmp[mic_permute[ib]]=mictimf_int[pb]*mic_win[ia];
          if(abs(mictimf_int[pb]) > tx_ad_maxamp)
            {
            tx_ad_maxamp=abs(mictimf_int[pb]);
            }
          ib--; 
          pa=(pa+2)&mictimf_mask;
          pb=(pb+mictimf_mask-1)&mictimf_mask;
          }
        }
// The microphone signal should be maximum 0x7fff or 0x7fffffffffff
// depending on whether 16 or 32 bit format is selected. 
      mictimf_px=(mictimf_px+mictimf_block)&mictimf_mask;
      fft_real_to_hermitian( mic_tmp, twosiz, micn, mic_table);
// Output is {Re(z^[0]),...,Re(z^[n/2),Im(z^[n/2-1]),...,Im(z^[1]).
// Clear those parts of the spectrum that we do not want and
// transfer the rest of the points to micfft while multiplying
// with our filter function. 
// Place the spectrum center at frequency zero so we can use
// the lowest possible sampling rate in further processing steps.
      z=&micfft[micfft_pa];
      pc=2*(tx_highest_bin/2-tx_lowest_bin/2);
      k=micfft_block-pc-2;
      while(pc < k)
        {
        z[pc  ]=0;
        z[pc+1]=0;
        pc+=2;
        }
      k=tx_lowest_bin;
      pc=2*tx_filter_ia1;
      while(pc < 2*mic_fftsize)
        {
        z[pc+1]=mic_tmp[k  ]*mic_filter[k];
        z[pc  ]=mic_tmp[twosiz-k]*mic_filter[k];
        pc+=2;
        k++;
        }
      pc=0;    
      while(k <= tx_highest_bin)
        {
        z[pc+1]=mic_tmp[k  ]*mic_filter[k];
        z[pc  ]=mic_tmp[twosiz-k]*mic_filter[k];
        pc+=2;
        k++;
        }
      micfft_pa=(micfft_pa+micfft_block)&micfft_mask;
      }
    break;

    case MODE_WCW:
    case MODE_NCW:
    case MODE_HSMS:
    case MODE_QRSS:
    switch(tx_input_mode)
      {
      case 0:
      while(mictimf_pa != mictimf_px)
        {
        t1=0;
        for(i=0; i<tx_pre_resample; i++)
          {
          t1+=abs(mictimf_shi[mictimf_px]-mictimf_shi[pb]);
          pb=mictimf_px;
          mictimf_px++;
          }
        mictimf_px=mictimf_px&mictimf_mask;
        tx_mic_key[mic_key_pa]=t1/(tx_pre_resample*1000);
        mic_key_pa=(mic_key_pa+1)&mic_key_mask;
        }
      break;

      case 1:
// In case there are two channels for the microphone, just use one of them!!
      while(mictimf_pa != mictimf_px)
        {
        t1=0;
        for(i=0; i<tx_pre_resample; i++)
          {
          t1+=abs(mictimf_shi[mictimf_px]-mictimf_shi[pb]);
          pb=mictimf_px;
          mictimf_px+=2;
          }
        mictimf_px=mictimf_px&mictimf_mask;
        tx_mic_key[mic_key_pa]=t1/(tx_pre_resample*1000);
        mic_key_pa=(mic_key_pa+1)&mic_key_mask;
        }
      break;
        
      case 2:
      while(mictimf_pa != mictimf_px)
        {
        t1=0;
        for(i=0; i<tx_pre_resample; i++)
          {
          t1+=abs(mictimf_int[mictimf_px]-mictimf_int[pb]);
          pb=mictimf_px;
          mictimf_px++;
          }
        mictimf_px=mictimf_px&mictimf_mask;
        tx_mic_key[mic_key_pa]=(t1/tx_pre_resample)/(0x7fff*1000);
        mic_key_pa=(mic_key_pa+1)&mic_key_mask;
        }
      break;
        
      case 3:
      while(mictimf_pa != mictimf_px)
        {
        t1=0;
        for(i=0; i<tx_pre_resample; i++)
          {
          t1+=abs(mictimf_int[mictimf_px]-mictimf_int[pb]);
          pb=mictimf_px;
          mictimf_px+=2;
          }
        mictimf_px=mictimf_px&mictimf_mask;
        tx_mic_key[mic_key_pa]=t1/(tx_pre_resample*1000);
        mic_key_pa=(mic_key_pa+1)&mic_key_mask;
        }
      break;
      }
    }  
  lir_sem_post(SEM_TXINPUT);
  lir_sched_yield();
  }
tx_input_x:;
thread_status_flag[THREAD_TX_INPUT]=THRFLAG_RETURNED;
lir_sem_post(SEM_TXINPUT);
while(thread_command_flag[THREAD_TX_INPUT] != THRFLAG_NOT_ACTIVE)
  {
  lir_sleep(1000);
  }
}



void init_txmem_spproc(void)
{
int i, k;
float t1;
// Speech processing is done in both the time domain and in the
// frequency domain.
// The lowest frequency of interest is 150 Hz so it will be
// sufficient to use FFTs that span 10 milliseconds.
// With sin squared windows we use 50% overlap and make a read
// every 5 millisecond. 
// There is a factor of two because the soundcard samples a real-
// valued signal but micsize is for complex transforms.
micsize=ui.tx_ad_speed*.01;
make_power_of_two(&micsize);
micn=0;
i=micsize;
while(i>1)
  {
  i>>=1;
  micn++;
  }
mic_sizhalf=micsize/2;
mictimf_block=ui.tx_ad_channels*micsize;
// Allow the input buffer to hold 40 milliseconds of data (or more)
mictimf_bufsiz=8*mictimf_block;
mictimf_mask=mictimf_bufsiz-1;
tx_read_bytes=ui.tx_ad_bytes*mictimf_block;
txad_hz_per_bin=(0.5*ui.tx_ad_speed)/micsize;
tx_lowest_bin=txssb.minfreq/txad_hz_per_bin;
if(tx_lowest_bin < 1)tx_lowest_bin=1;
if(tx_lowest_bin > micsize/3)tx_lowest_bin=micsize/3;
tx_highest_bin=txssb.maxfreq/txad_hz_per_bin;
if(tx_highest_bin > micsize/2-2)tx_highest_bin=micsize/2-2;
if(tx_highest_bin < tx_lowest_bin+1)tx_highest_bin=tx_lowest_bin+1;
t1=(tx_highest_bin-tx_lowest_bin+1)*txad_hz_per_bin;
if(ui.tx_da_channels == 1)t1*=2;
if(t1 > 0.75*ui.tx_da_speed)
  {
  t1=ui.tx_da_speed;
  if(ui.tx_da_channels == 1)t1/=2;
  t1/=txad_hz_per_bin;
  tx_highest_bin=tx_lowest_bin+t1;
  }
mic_fftsize=micsize;
mic_fftn=micn;
k=1.2*(tx_highest_bin-tx_lowest_bin);
k*=2;
while(k < mic_fftsize)
  {
  mic_fftsize/=2;
  mic_fftn--;
  }    
mic_sizhalf=mic_fftsize/2;
micfft_block=2*mic_fftsize;
micfft_bufsiz=8*mic_fftsize;
micfft_mask=micfft_bufsiz-1;
tx_filter_ia1=mic_fftsize+(tx_lowest_bin/2-tx_highest_bin/2-1);
tx_filter_ib1=tx_filter_ia1+tx_highest_bin-tx_lowest_bin-mic_fftsize;
k=mic_fftsize/tx_filter_ib1;
alc_fftn=mic_fftn;
alc_fftsize=mic_fftsize;
tx_pre_resample=1;
k++;
while(k<8)
  {
  tx_pre_resample*=2;
  k*=2;
  alc_fftn++;
  alc_fftsize*=2;
  }
alc_bufsiz=16*alc_fftsize;
alc_mask=alc_bufsiz-1;
alc_sizhalf=alc_fftsize/2;      
alc_block=2*alc_fftsize;
tx_resample_ratio=(float)(2*ui.tx_da_speed)/ui.tx_ad_speed;
tx_output_upsample=1;
txout_fftsize=micsize;
txout_fftn=micn;
while(tx_resample_ratio > 1.5)
  {
  tx_resample_ratio/=2;
  tx_output_upsample*=2;
  txout_fftsize*=2;
  txout_fftn++;
  }
new_tx_resample_ratio=tx_resample_ratio;
txout_sizhalf=txout_fftsize/2;
init_memalloc(txmem, MAX_TXMEM_ARRAYS);
if(ui.tx_ad_bytes == 2)
  {
  mem(1,&mictimf_shi,mictimf_bufsiz*sizeof(short int),0);
  }
else
  {
  mem(1,&mictimf_int,mictimf_bufsiz*sizeof(int),0);
  }
mem(2,&micfft,micfft_bufsiz*sizeof(float),0);
mem(3,&cliptimf,micfft_bufsiz*sizeof(float),0);
mem(4,&mic_tmp,2*micsize*sizeof(float),0);
mem(5,&mic_table,micsize*sizeof(COSIN_TABLE),0);
mem(6,&mic_permute,2*micsize*sizeof(short int),0);
mem(7,&mic_win,(1+micsize)*sizeof(float),0);
mem(8,&mic_filter,micsize*sizeof(float),0);
mem(9,&micfft_table,micsize*sizeof(COSIN_TABLE)/2,0);
mem(10,&micfft_permute,micsize*sizeof(short int),0);
mem(11,&micfft_win,(1+mic_sizhalf)*sizeof(float),0);
mem(12,&clipfft,alc_bufsiz*sizeof(float),0);
mem(13,&alctimf,alc_bufsiz*sizeof(float),0);
mem(14,&alctimf_pwrf,alc_bufsiz*sizeof(float)/2,0);
mem(15,&alctimf_pwrd,alc_bufsiz*sizeof(float)/2,0);
mem(16,&tx_resamp,2*txout_fftsize*sizeof(float),0);
mem(17,&resamp_tmp,alc_fftsize*sizeof(float),0);
mem(18,&alc_permute,alc_fftsize*sizeof(short int),0);
mem(19,&alc_table,alc_sizhalf*sizeof(COSIN_TABLE),0);
mem(20,&alc_win,(1+alc_sizhalf)*sizeof(float),0);
mem(21,&txout,2*txout_fftsize*sizeof(float),0);
mem(22,&txout_permute,txout_fftsize*sizeof(short int),0);
mem(23,&txout_table,txout_sizhalf*sizeof(COSIN_TABLE),0);
mem(24,&txout_tmp,txout_fftsize*sizeof(float),0);
if(ui.tx_da_bytes==4)
  {
  mem(25,&tx_out_int,ui.tx_da_channels*txout_fftsize*sizeof(int)/2,0);
  }
else
  {  
  mem(25,&tx_out_shi,ui.tx_da_channels*txout_fftsize*sizeof(int)/2,0);
  }
mem(26,&cliptimf_mute,micfft_bufsiz*sizeof(char)/mic_fftsize,0);  
mem(27,&clipfft_mute,alc_bufsiz*sizeof(char)/alc_block,0);  
mem(28,&alctimf_mute,alc_bufsiz*sizeof(char)/alc_fftsize,0);  
txmem_size=memalloc((int*)(&txmem_handle),"txmem");
if(txmem_size == 0)
  {
  lirerr(1261);
  return;
  }
make_permute(2, micn, micsize, mic_permute);
make_window(2,micsize, 2, mic_win);
if(ui.tx_ad_bytes != 2)
  {
  for(i=0; i<micsize; i++)mic_win[i]/=0x10000;
  }
t1=1/(1.62*0x7fff*mic_sizhalf);  
for(i=0; i<micsize; i++)mic_win[i]*=t1;
make_sincos(2, micsize, mic_table); 
init_fft(0,mic_fftn, mic_fftsize, micfft_table, micfft_permute);
make_window(2,mic_sizhalf, 2, micfft_win);
micfft_winfac=1.62*mic_fftsize;
t1=1/micfft_winfac;
for(i=0; i<mic_sizhalf; i++)
  {
  micfft_win[i]*=t1;
  }
for(i=0; i<micfft_bufsiz; i++)
  {
  micfft[i]=0;
  cliptimf[i]=0;
  }
init_fft(0,alc_fftn, alc_fftsize, alc_table, alc_permute);
make_window(2,alc_sizhalf, 2, alc_win);
t1=1/(1.62*alc_fftsize);
for(i=0; i<alc_sizhalf; i++)
  {
  alc_win[i]*=t1;
  }
for(i=0; i<alc_bufsiz; i++)
  {
  alctimf[i]=0;
  clipfft[i]=0;
  }
for(i=0; i<alc_bufsiz/2; i++)alctimf_pwrf[i]=0;
init_fft(0,txout_fftn, txout_fftsize, txout_table, txout_permute);
for(i=0; i<2*txout_fftsize; i++)tx_resamp[i]=0;

tx_filter_ia1=mic_fftsize+(tx_lowest_bin/2-tx_highest_bin/2-1);
tx_filter_ia2=alc_fftsize+(tx_lowest_bin/2-tx_highest_bin/2-1);
tx_filter_ia3=txout_fftsize+(tx_lowest_bin/2-tx_highest_bin/2-1);
tx_filter_ib1=tx_filter_ia1+tx_highest_bin-tx_lowest_bin-mic_fftsize;
tx_filter_ib2=tx_filter_ia2+tx_highest_bin-tx_lowest_bin-alc_fftsize;
tx_filter_ib3=tx_filter_ia3+tx_highest_bin-tx_lowest_bin-txout_fftsize;
tx_ssbproc_fqshift=(float)(tx_highest_bin-tx_filter_ib1)*txad_hz_per_bin;
// Construct the microphone input filter from the
// parameters supplied by the user.
for(i=0; i<mic_fftsize; i++)mic_filter[i]=0;
// Store the desired frequency response in dB.
t1=0;
for(i=tx_lowest_bin; i<=tx_highest_bin; i++)
  {
  mic_filter[i]=t1;
  t1+=0.001*txssb.slope*txad_hz_per_bin;
  }
k=(tx_highest_bin-tx_lowest_bin)/4;
t1=0;
for(i=tx_lowest_bin+k; i>=tx_lowest_bin; i--)
  {
  t1+=0.001*txssb.bass*txad_hz_per_bin;
  mic_filter[i]+=t1;
  }
t1=0;
for(i=tx_highest_bin-k; i<=tx_highest_bin; i++)
  {
  t1+=0.001*txssb.treble*txad_hz_per_bin;
  mic_filter[i]+=t1;
  }
// Convert from dB to linear scale (amplitude) and normalize
// the filter curve.
t1=0;
for(i=tx_lowest_bin; i<=tx_highest_bin; i++)
  {
  mic_filter[i]=pow(10.,0.05*mic_filter[i]);
  t1+=mic_filter[i]*mic_filter[i];
  }
t1=1/sqrt(t1/(tx_highest_bin-tx_lowest_bin+1));
for(i=tx_lowest_bin; i<=tx_highest_bin; i++)
  {
  mic_filter[i]*=t1;
  }
mic_filter[tx_lowest_bin]*=0.5;
mic_filter[tx_highest_bin]*=0.5;

micfft_bin_minpower_ref=pow(10.,0.1*(txssb.mic_f_threshold-80));
micfft_minpower_ref=pow(10.,0.1*(txssb.mic_t_threshold+txssb.mic_f_threshold-70));
tx_agc_factor=1;
tx_agc_decay=pow(2.718,-100.0/(txssb.mic_agc_time*txad_hz_per_bin));
txpwr_decay=pow(2.718,-1/(ui.tx_ad_speed*0.025));
tx_forwardpwr=0;
tx_backpwr=0;
tx_ph1=1;
tx_wrblock_no=0;
tx_wttim_ref=0;
memcheck(1,txmem,(int*)&txmem_handle);
make_tx_modes();
//lir_sleep(3000);  
tx_output_totsamp=0;
tx_output_totsamp=-lir_tx_output_samples();
tx_output_phstep=PI_L/6;
tx_output_phstep=0.25*PI_L;
tx_output_phstep=2*PI_L*(tg.freq*1000000+tx_ssbproc_fqshift)/ui.tx_da_speed;
tx_output_phase=0;
tx_resamp_ampfac1=1;
tx_resamp_ampfac2=1;
micfft_pa=0;
micfft_px=0;
mictimf_pa=0;
mictimf_px=0;
cliptimf_pa=0;
cliptimf_px=0;
clipfft_pa=0;
clipfft_px=0;
alctimf_pa=0;
alctimf_pb=0;
alctimf_fx=0;
tx_output_flag=0;
}


void make_txproc_filename(void)
{
switch (rx_mode)
  {
  case MODE_SSB:
  sprintf(&ssbproc_filename[11],"%02d",tg.spproc_no);
  txproc_filename=ssbproc_filename;
  break;

  case MODE_WCW:
  case MODE_NCW:
  case MODE_HSMS:
  case MODE_QRSS:
  sprintf(&cwproc_filename[10],"%02d",tg.spproc_no);
  txproc_filename=cwproc_filename;
  break;
  
  default:
  txproc_filename=NULL;
  break;
  }
}



int tx_ssb_step2(float *totpwr)
{
int i, k;
float t1, t2;
// ************* Step 2. *************
// Compute the power for each bin within the passband.
// Apply muting in the frequency domain by setting bins with too small
// signal power to zero.
// Collect the sum of the powers from surviving bins and mute the entire
// block in case the total power is below threshold.
// Here we use the agc factor as determined from the the
// previous block and attenuated with the decay factor.
// Note that the agc factor is limited to 20 dB at this point.
// A big pulse will not kill the signal for a long time!!
t1=0;
k=0;
for(i=tx_filter_ia1; i<mic_fftsize; i++)
  {
  t2=micfft[micfft_px+2*i  ]*micfft[micfft_px+2*i  ]+
     micfft[micfft_px+2*i+1]*micfft[micfft_px+2*i+1];
  if(t2 > micfft_bin_minpower)
    {
    k++;
    t1+=t2;
    }
  else
    {
    micfft[micfft_px+2*i  ]=0;
    micfft[micfft_px+2*i+1]=0;
    }
  }
for(i=0; i<=tx_filter_ib1; i++)
  {
  t2=micfft[micfft_px+2*i  ]*micfft[micfft_px+2*i  ]+
     micfft[micfft_px+2*i+1]*micfft[micfft_px+2*i+1];
  if(t2 > micfft_bin_minpower)
    {
    k++;
    t1+=t2;
    }
  else
    {
    micfft[micfft_px+2*i  ]=0;
    micfft[micfft_px+2*i+1]=0;
    }
  }
if(k>0 && t1 < micfft_minpower)
  {
  for(i=tx_filter_ia1; i<mic_fftsize; i++)
    {
    micfft[micfft_px+2*i  ]=0;
    micfft[micfft_px+2*i+1]=0;
    }
  for(i=0; i<=tx_filter_ib1; i++)
    {
    micfft[micfft_px+2*i  ]=0;
    micfft[micfft_px+2*i+1]=0;
    }
  k=0;
  }
totpwr[0]=t1;    
return k;
}

void tx_ssb_step3(float *totamp)
{
int i;
float t1, t2, r1;
// ************* Step 3. *************
// Find out what the average amplitude would become if the old AGC
// factor would be applied. Change the agc factor in case the
// average amplitude is above the limit.
// Note that this step is only executed if k!=0.
// When k==0, the entire block is muted. The operation
// would be meaningless on a zero transform.
t1=sqrt(totamp[0]);
r1=pow(10.0,0.05*txssb.mic_gain);
t2=r1*t1*tx_agc_factor*(tx_highest_bin-tx_lowest_bin+1);
if(t2 > TX_AGC_THRESHOLD*(tx_highest_bin-tx_lowest_bin+1))
  {
  tx_agc_factor=TX_AGC_THRESHOLD/(r1*t1);
  }
t2=tx_agc_factor*(tx_highest_bin-tx_lowest_bin+1)*r1;
t2*=pow(10.0,0.05*(txssb.mic_out_gain-60));
// Apply the AGC within the current fft block.
for(i=tx_filter_ia1; i<mic_fftsize; i++)
  {
  micfft[micfft_px+2*i  ]*=t2;
  micfft[micfft_px+2*i+1]*=t2;
  }
for(i=0; i<=tx_filter_ib1; i++)
  {
  micfft[micfft_px+2*i  ]*=t2;
  micfft[micfft_px+2*i+1]*=t2;
  }
totamp[0]=t1;
}

void tx_ssb_step4(float ampinv,float *t1, float *prat)
{
int i, k, panew;
char mute_current, mute_previous;
float r1, r2, t2;
prat[0]=0;
t1[0]=0;
if( (tx_filter_ia1&1) != 0)tx_ph1*=-1;
panew=(cliptimf_pa+mic_fftsize)&micfft_mask;
if(ampinv == 0)
  {
  mute_current=TRUE;
  }
else
  {  
  mute_current=FALSE;
  }
cliptimf_mute[cliptimf_pa/mic_fftsize]=mute_current;
mute_previous=cliptimf_mute[((cliptimf_pa-mic_fftsize+micfft_bufsiz)&
                           micfft_mask)/mic_fftsize];
// We do nothing in case both the current transform and the
// previous one are muted.
if( (mute_current&mute_previous) == FALSE)
  {  
  if( (mute_current|mute_previous) == FALSE)
    {
// None of the transforms is muted. 
// Do normal (continous) back transformation.
    r1=pow(10.0,0.05*txssb.rf1_gain)/TX_AGC_THRESHOLD;
    t2=0;
    r1*=ampinv;
    for(i=0; i<mic_fftsize; i+=2)
      {
      cliptimf[cliptimf_pa+i  ]=tx_ph1*
                      (cliptimf[cliptimf_pa+i  ]+r1*micfft[micfft_px+i  ]);
      cliptimf[cliptimf_pa+i+1]=tx_ph1*
                      (cliptimf[cliptimf_pa+i+1]+r1*micfft[micfft_px+i+1]);
      t2=cliptimf[cliptimf_pa+i  ]*cliptimf[cliptimf_pa+i  ]+
         cliptimf[cliptimf_pa+i+1]*cliptimf[cliptimf_pa+i+1];
      prat[0]+=t2;
      if(t2>t1[0])t1[0]=t2;
      if(t2 > 1)
        {
        t2=1/sqrt(t2);
        cliptimf[cliptimf_pa+i  ]*=t2;
        cliptimf[cliptimf_pa+i+1]*=t2;
        }
      }
    if( (tx_filter_ia1&1) != 0)r1*=-1;
    for(i=0; i<mic_fftsize; i++)
      {
      cliptimf[panew+i]=r1*micfft[micfft_px+mic_fftsize+i];
      }
    }
  else
    {
// This is the first or the last transform.
// Multiply the beginning and the end with a sine squared window
// to make the start and end filtered with a sine to power 4
// window in total.    
// The first and last transforms are likely to be incorrect
// because muting in the frequency domain will distort the
// signal if only the very strongest bins have survived
// the muting process. This is particularly obvious if the
// microphone signal is a keyed tone which is present only
// a short time in the last or first transform. In such a
// situation the signal would be extended over the entire
// transform in case only a single bin would survive.
// The added sine squared window will remove the abrupt
// start that the signal otherwise would have.
    if(mute_current)
      {
      k=0;
      for(i=mic_fftsize-2; i>=0; i-=2)
        {
        r1=micfft_winfac*micfft_win[k];
        cliptimf[cliptimf_pa+i  ]=tx_ph1*cliptimf[cliptimf_pa+i  ]*r1;
        cliptimf[cliptimf_pa+i+1]=tx_ph1*cliptimf[cliptimf_pa+i+1]*r1;
        k++;
        t2=cliptimf[cliptimf_pa+i  ]*cliptimf[cliptimf_pa+i  ]+
           cliptimf[cliptimf_pa+i+1]*cliptimf[cliptimf_pa+i+1];
        if(t1[0] < MAX_DYNRANGE && t2 < MAX_DYNRANGE)
          {
          cliptimf[cliptimf_pa+i  ]=0;
          cliptimf[cliptimf_pa+i+1]=0;
          }
        else
          {
          prat[0]+=t2;
          if(t2>t1[0])t1[0]=t2;
          if(t2 > 1)
            {
            t2=1/sqrt(t2);
            cliptimf[cliptimf_pa+i  ]*=t2;
            cliptimf[cliptimf_pa+i+1]*=t2;
        t2=cliptimf[cliptimf_pa+i  ]*cliptimf[cliptimf_pa+i  ]+
           cliptimf[cliptimf_pa+i+1]*cliptimf[cliptimf_pa+i+1];
            }
          }
        }
      for(i=0; i<mic_fftsize; i++)
        {
        cliptimf[panew+i]=0;
        }
      }
    else // if(mute_previous)
      {
      r1=pow(10.0,0.05*txssb.rf1_gain)/TX_AGC_THRESHOLD;
      t2=0;
      r1*=ampinv;
      for(i=0; i<mic_fftsize; i+=2)
        {
        r2=r1*micfft_winfac*micfft_win[i>>1];
        cliptimf[cliptimf_pa+i  ]=tx_ph1*r2*micfft[micfft_px+i  ];
        cliptimf[cliptimf_pa+i+1]=tx_ph1*r2*micfft[micfft_px+i+1];
        t2=cliptimf[cliptimf_pa+i  ]*cliptimf[cliptimf_pa+i  ]+
           cliptimf[cliptimf_pa+i+1]*cliptimf[cliptimf_pa+i+1];
        if(t1[0] < MAX_DYNRANGE && t2 < MAX_DYNRANGE)
          {
          cliptimf[cliptimf_pa+i  ]=0;
          cliptimf[cliptimf_pa+i+1]=0;
          }
        else
          {
          prat[0]+=t2;
          if(t2>t1[0])t1[0]=t2;
          if(t2 > 1)
            {
            t2=1/sqrt(t2);
            cliptimf[cliptimf_pa+i  ]*=t2;
            cliptimf[cliptimf_pa+i+1]*=t2;
            }
          }
        }
      if( (tx_filter_ia1&1) != 0)r1*=-1;
      for(i=0; i<mic_fftsize; i++)
        {
        cliptimf[panew+i]=r1*micfft[micfft_px+mic_fftsize+i];
        }
      }
    }
  }
cliptimf_pa=panew;
micfft_px=(micfft_px+micfft_block)&micfft_mask;
}

void tx_ssb_step5(void)
{
int i, j, pb, nx ,nb;
while( ((cliptimf_pa-cliptimf_px+micfft_bufsiz)&micfft_mask) > 
                                                           2*mic_fftsize)
  {                                                  
  pb=(cliptimf_px+2*(mic_fftsize-1))&micfft_mask;
  nx=cliptimf_px/mic_fftsize;
  nb=((cliptimf_px+mic_fftsize)&micfft_mask)/mic_fftsize;
  if(cliptimf_mute[nx]==FALSE/* && cliptimf_mute[nb]==FALSE*/)
    {
    j=mic_fftsize-1;
    for(i=0; i<mic_sizhalf; i++)
      {
      clipfft[clipfft_pa+2*i  ]=cliptimf[cliptimf_px+2*i  ]*micfft_win[i]; 
      clipfft[clipfft_pa+2*i+1]=cliptimf[cliptimf_px+2*i+1]*micfft_win[i]; 
      clipfft[clipfft_pa+2*j  ]=cliptimf[pb  ]*micfft_win[i]; 
      clipfft[clipfft_pa+2*j+1]=cliptimf[pb+1]*micfft_win[i]; 
      j--;
      pb-=2;
      }      
    fftforward(mic_fftsize, mic_fftn, &clipfft[clipfft_pa], 
                                               micfft_table, micfft_permute);
    if(tx_setup_flag == TRUE)
      {
      lir_sched_yield();
      if(screen_count>=MAX_SCREENCOUNT)
        {
        show_txfft(&clipfft[clipfft_pa],0,1,mic_fftsize);
        }
      }
    for(i=tx_filter_ib1+1; i<tx_filter_ia1; i++)
      {
      clipfft[clipfft_pa+2*i  ]=0;
      clipfft[clipfft_pa+2*i+1]=0;
      }
    clipfft_mute[clipfft_pa/alc_block]=FALSE;
    }
  else
    {
    clipfft_mute[clipfft_pa/alc_block]=TRUE;
    }      
  cliptimf_px=(cliptimf_px+mic_fftsize)&micfft_mask;
  clipfft_pa=(clipfft_pa+alc_block)&alc_mask;
  }
}

float tx_ssb_step6(float *prat)
{
int i, k, pb, mute_previous;
float t1, t2, pmax;
mute_previous=clipfft_mute[((clipfft_px-alc_block+alc_bufsiz)
                                                     &alc_mask)/alc_block];
pmax=0;
prat[0]=0;
if(clipfft_mute[clipfft_px/alc_block] == FALSE)
  {
  alctimf_mute[alctimf_pa/alc_fftsize]=FALSE;
  if(mic_fftsize < alc_fftsize)
    {
    i=tx_filter_ia1;
    k=tx_filter_ia2;
    while(i<mic_fftsize)
      {
      clipfft[clipfft_px+2*k  ]=clipfft[clipfft_px+2*i  ];
      clipfft[clipfft_px+2*k+1]=clipfft[clipfft_px+2*i+1];
      i++;
      k++;
      }      
    }  
  for(i=tx_filter_ib1+1; i<tx_filter_ia2; i++)
    {
    clipfft[clipfft_px+2*i]=0;
    clipfft[clipfft_px+2*i+1]=0;
    }      
  fftback(alc_fftsize, alc_fftn, &clipfft[clipfft_px], 
                                              alc_table, alc_permute);
  if(mute_previous) 
    {
    for(i=0; i<alc_sizhalf; i++)
      {
      alctimf[alctimf_pa+2*i  ]=clipfft[clipfft_px+2*i  ];
      alctimf[alctimf_pa+2*i+1]=clipfft[clipfft_px+2*i+1];
      t2=alctimf[alctimf_pa+2*i  ]*alctimf[alctimf_pa+2*i  ]+
         alctimf[alctimf_pa+2*i+1]*alctimf[alctimf_pa+2*i+1];
      prat[0]+=t2;
      if(t2>pmax)pmax=t2;
      tx_forwardpwr*=txpwr_decay;
      if(tx_forwardpwr<t2)tx_forwardpwr=t2;
      alctimf_pwrf[(alctimf_pa>>1)+i]=tx_forwardpwr;
      }
    }
  else
    {
    for(i=0; i<alc_sizhalf; i++)
      {
      alctimf[alctimf_pa+2*i  ]+=clipfft[clipfft_px+2*i  ];
      alctimf[alctimf_pa+2*i+1]+=clipfft[clipfft_px+2*i+1];
      t2=alctimf[alctimf_pa+2*i  ]*alctimf[alctimf_pa+2*i  ]+
         alctimf[alctimf_pa+2*i+1]*alctimf[alctimf_pa+2*i+1];
      prat[0]+=t2;
      if(t2>pmax)pmax=t2;
      tx_forwardpwr*=txpwr_decay;
      if(tx_forwardpwr<t2)tx_forwardpwr=t2;
      alctimf_pwrf[(alctimf_pa>>1)+i]=tx_forwardpwr;
      }
    }
  }
else
  {
  if(mute_previous)
    {
    tx_forwardpwr=0;
    alctimf_mute[alctimf_pa/alc_fftsize]=TRUE;
    }
  else
    {
    alctimf_mute[alctimf_pa/alc_fftsize]=FALSE;
    for(i=0; i<alc_sizhalf; i++)
      {
      t2=alctimf[alctimf_pa+2*i  ]*alctimf[alctimf_pa+2*i  ]+
         alctimf[alctimf_pa+2*i+1]*alctimf[alctimf_pa+2*i+1];
      prat[0]+=t2;
      if(t2>pmax)pmax=t2;
      tx_forwardpwr*=txpwr_decay;
      if(tx_forwardpwr<t2)tx_forwardpwr=t2;
      alctimf_pwrf[(alctimf_pa>>1)+i]=tx_forwardpwr;
      }
    }
  }  
if( alctimf_mute[alctimf_pa/alc_fftsize] == FALSE)
  {
// Store an exponential power fall-off with the same time constant, 50 ms,
// in the reverse direction.
  i=alc_sizhalf-1;
  t1=0;
  while(i >= 0)
    {
    t1*=txpwr_decay;
    if(t1 < alctimf_pwrf[(alctimf_pa>>1)+i])t1=alctimf_pwrf[(alctimf_pa>>1)+i];
    alctimf_pwrd[(alctimf_pa>>1)+i]=t1;
    i--;
    }
  pb=alctimf_pa;
  t1=alctimf_pwrd[(pb>>1)];
  i=0;
  while( pb != ((int)(alctimf_fx+4+alc_bufsiz)&alc_mask) && i == 0)
    {
    t1*=txpwr_decay;
    pb=(pb+alc_mask)&alc_mask;
    if(t1 > alctimf_pwrf[(pb>>1)])
      {
      alctimf_pwrd[(pb>>1)]=t1;
      }
    else
      {
      i=1;
      }  
    }
  }  
alctimf_pa=(alctimf_pa+alc_fftsize)&alc_mask;
if(clipfft_mute[clipfft_px/alc_block] == FALSE)
  {
  for(i=0; i<alc_fftsize; i++)
    {
    alctimf[alctimf_pa+i]=clipfft[clipfft_px+alc_fftsize+i];
    }
  }
else
  {
  for(i=0; i<alc_fftsize; i++)
    {
    alctimf[alctimf_pa+i]=0;
    }
  }
clipfft_px=(clipfft_px+alc_block)&alc_mask;
return pmax;
}

float tx_ssb_step7(float *prat)
{
int i;
float t1, t2, pmax;
if(alctimf_mute[alctimf_pb/alc_fftsize]==FALSE)
  {
  pmax=0;
  prat[0]=0;
  for(i=0; i<alc_sizhalf; i++)
    {
    t1=alctimf_pwrd[(alctimf_pb>>1)+i];    
    if(t1 > 1)
      {
      t1=1/sqrt(t1);
      alctimf[alctimf_pb+2*i  ]*=t1;
      alctimf[alctimf_pb+2*i+1]*=t1;
      }
    t2=alctimf[alctimf_pb+2*i  ]*alctimf[alctimf_pb+2*i  ]+
       alctimf[alctimf_pb+2*i+1]*alctimf[alctimf_pb+2*i+1];
// We do not want more than MAX_DYNRANGE dynamic range.
// any signal below is just rounding errors in muted periods.
    if(t2 < MAX_DYNRANGE)
      {  
      alctimf[alctimf_pb+2*i  ]=0;
      alctimf[alctimf_pb+2*i+1]=0;
      }
    if(t2 > pmax)pmax=t2;
    prat[0]+=t2;
    }
  }
else
  {
  pmax=0;
  prat[0]=0;
  for(i=0; i<alc_fftsize; i++)
    {
    alctimf[alctimf_pb+i]=0;
    }
  }
alctimf_pb=(alctimf_pb+alc_fftsize)&alc_mask;
return pmax;
}

void tx_ssb_step8(void)
{
if(tx_output_flag == 0)
  {
  if( ((alctimf_pb-(int)(alctimf_fx)+alc_bufsiz)&alc_mask) > 4*alc_fftsize)
    {
    tx_output_flag=1;
    tx_resamp_pa=0;
    tx_resamp_px=0;
    }
  }  
if(tx_output_flag == 1)
  {
  resample_tx_output();
  }
}

void spproc_setup(void)
{
char s[80];
int i, k;
int setup_mode;
int ad_pix1;
int ad_pix2;
int mute_pix1;
int mute_pix2;
int micagc_pix1;
int micagc_pix2;
int rf1agc_pix1;
int rf1agc_pix2;
int rf1clip_pix1;
int rf1clip_pix2;
int alc_pix1;
int alc_pix2;
float t1, t2, r2;
float prat, pmax;
int old_display_ptr;
int old_admax[DISPLAY_HOLD_SIZE];
float old_rf1agc[DISPLAY_HOLD_SIZE];
float old_rf1clip[DISPLAY_HOLD_SIZE];
float old_prat1[DISPLAY_HOLD_SIZE];
float old_pmax1[DISPLAY_HOLD_SIZE];
float old_prat2[DISPLAY_HOLD_SIZE];
float old_pmax2[DISPLAY_HOLD_SIZE];
float old_prat3[DISPLAY_HOLD_SIZE];
float old_pmax3[DISPLAY_HOLD_SIZE];
int default_spproc_no;
rx_mode=MODE_SSB;
default_spproc_no=tg.spproc_no;
tx_setup_flag=TRUE;
setup_mode=TX_SETUP_AD;
restart:;
screen_count=0;
if(txssb.minfreq < SSB_MINFQ_LOW)txssb.minfreq=SSB_MINFQ_LOW;
if(txssb.minfreq > SSB_MINFQ_HIGH)txssb.minfreq=SSB_MINFQ_HIGH;
if(txssb.maxfreq < SSB_MAXFQ_LOW)txssb.maxfreq=SSB_MAXFQ_LOW;
if(txssb.maxfreq > SSB_MAXFQ_HIGH)txssb.maxfreq=SSB_MAXFQ_HIGH;
if(txssb.slope > SSB_MAXSLOPE)txssb.slope=SSB_MAXSLOPE;
if(txssb.slope < SSB_MINSLOPE)txssb.slope=SSB_MINSLOPE;
if(txssb.bass > SSB_MAXBASS)txssb.bass=SSB_MAXBASS;
if(txssb.bass < SSB_MINBASS)txssb.bass=SSB_MINBASS;
if(txssb.treble > SSB_MAXTREBLE)txssb.treble=SSB_MAXTREBLE;
if(txssb.treble < SSB_MINTREBLE)txssb.treble=SSB_MINTREBLE;
if(txssb.mic_f_threshold < 0)txssb.mic_f_threshold=0;
if(txssb.mic_f_threshold > SSB_MAX_MICF)txssb.mic_f_threshold=SSB_MAX_MICF;
if(txssb.mic_t_threshold < 0)txssb.mic_t_threshold=0;
if(txssb.mic_t_threshold > SSB_MAX_MICT)txssb.mic_t_threshold=SSB_MAX_MICT;
if(txssb.mic_gain < 0)txssb.mic_gain=0;
if(txssb.mic_gain > SSB_MAX_MICGAIN)txssb.mic_gain=SSB_MAX_MICGAIN;
if(txssb.mic_agc_time < 0)txssb.mic_agc_time=0;
if(txssb.mic_agc_time > SSB_MAX_MICAGC_TIME)
                                   txssb.mic_agc_time=SSB_MAX_MICAGC_TIME;
if(txssb.mic_out_gain < 0)txssb.mic_out_gain=0;
if(txssb.mic_out_gain > SSB_MAX_MICOUT_GAIN)
                                    txssb.mic_out_gain=SSB_MAX_MICOUT_GAIN;
if(txssb.rf1_gain < SSB_MIN_RF1_GAIN)txssb.rf1_gain=SSB_MIN_RF1_GAIN;
if(txssb.rf1_gain > SSB_MAX_RF1_GAIN)txssb.rf1_gain=SSB_MAX_RF1_GAIN;
open_tx_output();
open_tx_input();
if(kill_all_flag)return;
init_txmem_spproc();
if(kill_all_flag)return;
tx_show_siz=micsize;
if(tx_show_siz < alc_fftsize)tx_show_siz=alc_fftsize;
if(tx_show_siz < txout_fftsize)tx_show_siz=txout_fftsize;
txtrace=malloc(MAX_DISPLAY_TYPES*tx_show_siz*sizeof(short int));
if(txtrace == NULL)
  {
  lirerr(778543);
  return;
  }
lir_mutex_init();
for(i=0; i<MAX_DISPLAY_TYPES*tx_show_siz; i++)txtrace[i]=screen_height-1;
txtrace_gain=0.1;
txtrace_height=screen_height-text_height*(TXPAR_INPUT_LINE+1);
lir_sleep(50000);
linrad_thread_create(THREAD_TX_INPUT);
old_display_ptr=0;
for(i=0; i<DISPLAY_HOLD_SIZE; i++)
  {
  old_admax[i]=0;
  old_rf1agc[i]=0;
  old_rf1clip[i]=0;
  old_prat1[i]=90;
  old_pmax1[i]=0;
  old_prat2[i]=90;
  old_pmax2[i]=0;
  old_prat3[i]=90;
  old_pmax3[i]=0;
  }
clear_screen();
tx_ad_maxamp=0;
ad_pix1=0;
ad_pix2=0;
micagc_pix1=0;
micagc_pix2=0;
rf1agc_pix1=0;
rf1agc_pix2=0;
mute_pix1=0;
mute_pix2=0;
rf1clip_pix1=0;
rf1clip_pix2=0;
alc_pix1=0;
alc_pix2=0;
settextcolor(14);
make_txproc_filename();
sprintf(s,"Press 'S' to save %s, 'R' to read.  '+' or '-' to change file no.",
                                                 txproc_filename);
lir_text(0,TX_BAR_AD_LINE-2,s);
lir_text(0,TX_BAR_AD_LINE-1,"Arrow up/down to change spectrum scale.");
lir_text(0,TX_BAR_AD_LINE,"'M' =");
lir_text(0,TX_BAR_MUTE_LINE,"'Q' =");
lir_text(0,TX_BAR_MICAGC_LINE,"'A' =");
lir_text(0,TX_BAR_RF1AGC_LINE,"'D' =");
lir_text(0,TX_BAR_RF1CLIP_LINE,"'C' =");
settextcolor(7);
lir_text(6,TX_BAR_AD_LINE,"Soundcard");
lir_text(6,TX_BAR_MUTE_LINE,"Mute");
tx_bar(16,TX_BAR_AD_LINE,-1,-1);
tx_bar(16,TX_BAR_MUTE_LINE,-1,-1);
lir_text(6,TX_BAR_MICAGC_LINE,"MIC AGC");
tx_bar(16,TX_BAR_MICAGC_LINE,-1,-1);
lir_text(6,TX_BAR_RF1AGC_LINE,"RF1 AGC");
tx_bar(16,TX_BAR_RF1AGC_LINE,-1,-1);
lir_text(6,TX_BAR_RF1CLIP_LINE,"RF clip");
tx_bar(16,TX_BAR_RF1CLIP_LINE,-1,-1);
lir_text(6,TX_BAR_ALC_LINE,"ALC");
tx_bar(16,TX_BAR_ALC_LINE,-1,-1);
switch (setup_mode)
  {
  case TX_SETUP_AD:
  lir_text(0,0,
            "Verify AF clip level, (some device drivers give too few bits)");
  lir_text(0,1,"Use mixer program to set volume for the AF to never reach");
  lir_text(0,2,"its maximum during normal operation.");
  lir_text(0,3,
              "Press 'L' or 'H' to change cut-off frequencies,'F','B' or 'T'");
  lir_text(0,4,"for slope, bass or treble to change frequency response.");
  sprintf(s,
    "Low %.0f Hz  High %.0f Hz   Slope %d dB/kHz   Bass %d dB  Treble %d dB",
       tx_lowest_bin*txad_hz_per_bin, tx_highest_bin*txad_hz_per_bin,
                                            txssb.slope, txssb.bass, txssb.treble);
  lir_text(0,5,s);
  break;

  case TX_SETUP_MUTE:
  lir_text(0,0,"Press 'F' to set frequency domain threshold or");
  lir_text(0,1,"Press 'T' to set time domain threshold for signal");
  
  lir_text(0,2,"to become muted when only background noise is present.");
  sprintf(s,"Thresholds:  F=%d dB   T=%d dB",
                                 txssb.mic_f_threshold, txssb.mic_t_threshold);
  lir_text(0,3,s);
  break;
  
  case TX_SETUP_MICAGC:
  lir_text(0,0,"Press 'V' to set mic volume for suitable AGC action");
  lir_text(0,1,"or 'T' to set decay time constant");
  sprintf(s,"Vol=%d dB   T=%.2f sek",txssb.mic_gain, 0.01*txssb.mic_agc_time);
  lir_text(0,2,s);
  break;
  
  case TX_SETUP_RF1AGC:
  lir_text(0,0,"Press 'B' to set gain for some AGC action");
  sprintf(s,"Gain=%d dB",txssb.mic_out_gain);
  lir_text(0,2,s);
  break;

  case TX_SETUP_RF1CLIP:
  lir_text(0,0,"Press 'B' to set RF1 gain for desired clip level.");
  lir_text(0,1,"Bar range is 0 to 30 dB");
  sprintf(s,"RF1 gain=%d dB",txssb.rf1_gain);
  lir_text(0,2,s);
  break;
  }
while(!kill_all_flag)
  {
  if(screen_count>=MAX_SCREENCOUNT)screen_count=0;
  screen_count++;
  old_display_ptr++;
  if(old_display_ptr>=DISPLAY_HOLD_SIZE)old_display_ptr=0;
// Show the microphone input level. 
  ad_pix2=tx_ad_maxamp;
  if(ui.tx_ad_bytes != 2)
    {
    ad_pix2/=0x10000;
    }  
  old_admax[old_display_ptr]=ad_pix2;
  t1=0;
  for(i=0; i<DISPLAY_HOLD_SIZE; i++)
    {
    if(t1<old_admax[i])t1=old_admax[i];
    }
  t1=20*log10(t1/0x8000);  
  sprintf(s,"%.1fdB  ",t1);
  lir_text(18+TX_BAR_SIZE/text_width, TX_BAR_AD_LINE, s);
  ad_pix2=(ad_pix2*TX_BAR_SIZE)/0x8000;    
  tx_bar(16,TX_BAR_AD_LINE,ad_pix1,ad_pix2);
  ad_pix1=ad_pix2;
  tx_ad_maxamp=0.95*tx_ad_maxamp;
// **************************************************************
// The setup routine contains the same processing steps as the
// transmit routine, but it has varoius display functions added
// into it. The processing steps are numbered and described below.
// **************************************************************
//
// ************* Step 1. *************
// Wait for a new block of data from the mic input routine.
// Note that the data arrives as fourier transforms (micfft) and
// that the mic input routine tx_input() already has applied
// the filtering of the microphone signal.
  lir_sem_wait(SEM_TXINPUT);
  tx_agc_factor=tx_agc_decay*tx_agc_factor+(1-tx_agc_decay);
  micfft_bin_minpower=micfft_bin_minpower_ref*tx_agc_factor*tx_agc_factor;
  micfft_minpower=micfft_minpower_ref;//*tx_agc_factor*tx_agc_factor;
  if(screen_count>=MAX_SCREENCOUNT)
    {
    show_txfft(&micfft[micfft_px],micfft_bin_minpower,0,mic_fftsize);
    }
  k=tx_ssb_step2(&t1);
  if(k==0)
    {
    sprintf(s,"MUTED");
    }
  else
    {
    sprintf(s,"     ");
    tx_ssb_step3(&t1);
    }
  mute_pix2=(k*TX_BAR_SIZE)/(tx_highest_bin-tx_lowest_bin+1);    
  tx_bar(16,TX_BAR_MUTE_LINE,mute_pix1,mute_pix2);
  mute_pix1=mute_pix2;
  settextcolor(15);  
  lir_text(18+TX_BAR_SIZE/text_width,TX_BAR_MUTE_LINE,s);
  settextcolor(7);      
// In case AGC has reduced the gain by more than 20 dB, set the
// AGC factor to -20 dB immediately because voice should never
// be as loud. This is to avoid the agc release time constant
// for impulse sounds caused by hitting the microphone etc.       
  t1=20*log10(tx_agc_factor);  
  sprintf(s,"%.1fdB  ",t1);
  if(tx_agc_factor < 0.1)tx_agc_factor=0.1;
  lir_text(18+TX_BAR_SIZE/text_width, TX_BAR_MICAGC_LINE, s);
  micagc_pix2=-log10(tx_agc_factor)*TX_BAR_SIZE;
  tx_bar(16,TX_BAR_MICAGC_LINE,micagc_pix1,micagc_pix2);
  micagc_pix1=micagc_pix2;
// ************* Step 4. *************
// Go back to the time domain and store the signal in cliptimf
// Remember that we use sin squared windows and that
// the transforms overlap with 50%.
//
// Ideally, the peak amplitude should be 1.0, the audio AGC
// should be active and keep the power level nearly constant.
// For a voice signal the waveform will differ from time to time
// and therefore the peak to average power ratio will vary.
// Compute the peak amplitude for the current transform
// and save it for the next time we arrive here.
// Use the peak amplitude for an AGC in the time domain (Hilbert space).
// The Hilbert space AGC is a constant that may vary from one
// FFT block to the next one. It is equivalent with an AM modulator
// with a bandwidth corresponding to a single bin in the FFT so this
// AM modulation will not increase the bandwidth notably, but it
// will bring the RF amplitude near unity always.
//
// Finally, apply an amplitude limiter to the complex time domain signal.
// It will work as an RF limiter on the SSB signal and cause a lot of
// signal outside the passband.
  if(tx_agc_factor < 0.1)tx_agc_factor=0.1;
  if(k!=0)
    {
    fftback(mic_fftsize, mic_fftn, &micfft[micfft_px], 
                                              micfft_table, micfft_permute);
    lir_sched_yield();
    r2=0;
    for(i=0; i<mic_fftsize; i++)
      {
      t2=micfft[micfft_px+2*i  ]*micfft[micfft_px+2*i  ]+
         micfft[micfft_px+2*i+1]*micfft[micfft_px+2*i+1];
      if(t2>r2)r2=t2;
      }
    if(r2 > 0.01*MAX_DYNRANGE)
      {
      t1=sqrt(r2);
      r2=1/t1;
      old_rf1agc[old_display_ptr]=t1;
      if(t1>10)t1=10;
      rf1agc_pix2=log10(t1)*TX_BAR_SIZE;
      if(r2>10/tx_agc_factor)r2=10/tx_agc_factor;
      if(rf1agc_pix2 < 0)rf1agc_pix2=0;
      }
    else
      {
      r2=0;
      rf1agc_pix2=0;
      old_rf1agc[old_display_ptr]=0;
      }
    }          
  else
    {
    r2=0;
    rf1agc_pix2=0;
    old_rf1agc[old_display_ptr]=0;
    }
  t1=0;
  for(i=0; i<DISPLAY_HOLD_SIZE; i++)
    {
    if(t1<old_rf1agc[i])t1=old_rf1agc[i];
    }
  if(t1 > 0)t1=20*log10(t1);  
  sprintf(s,"%.1fdB  ",t1);
  lir_text(18+TX_BAR_SIZE/text_width, TX_BAR_RF1AGC_LINE, s);
  tx_bar(16,TX_BAR_RF1AGC_LINE,rf1agc_pix1,rf1agc_pix2);
  rf1agc_pix1=rf1agc_pix2;
  tx_ssb_step4(r2,&pmax,&prat);
  prat/=mic_sizhalf;
  old_prat1[old_display_ptr]=prat;
  old_pmax1[old_display_ptr]=pmax;
  prat=0;
  for(i=0; i<DISPLAY_HOLD_SIZE; i++)
    {
    prat+=old_prat1[i];
    }
  prat/=DISPLAY_HOLD_SIZE;  
  for(i=0; i<DISPLAY_HOLD_SIZE; i++)
    {
    if(pmax<old_pmax1[i])pmax=old_pmax1[i];
    }
  if(prat>0.000000001 && pmax > prat)
    {
    prat=10*log10(pmax/prat);  
    }
  else
    {
    prat=99;
    } 
  sprintf(s,"[%.1f]  ",prat);
  lir_text(28+TX_BAR_SIZE/text_width, TX_BAR_MICAGC_LINE, s);
  old_rf1clip[old_display_ptr]=pmax;
  if(pmax > 1000)pmax=1000;  
  if(pmax > 1)
    {  
    rf1clip_pix2=log10(pmax)*TX_BAR_SIZE/3;
    }
  else
    {
    rf1clip_pix2=0;
    }
  t1=0;
  for(i=0; i<DISPLAY_HOLD_SIZE; i++)
    {
    if(t1<old_rf1clip[i])t1=old_rf1clip[i];
    }
  if(t1 > 1)
    {
    t1=10*log10(t1);  
    }
  else
    {
    t1=0;
    }  
  sprintf(s,"%.1fdB  ",t1);
  lir_text(18+TX_BAR_SIZE/text_width, TX_BAR_RF1CLIP_LINE, s);
  tx_bar(16,TX_BAR_RF1CLIP_LINE,rf1clip_pix1,rf1clip_pix2);
  rf1clip_pix1=rf1clip_pix2;
// ************* Step 5. *************
// At this point we have applied an RF clipper by limiting power in
// Hilbert space. As a consequence we have produced intermodulation
// products outside the desired passband.
// Go to the frequency domain and remove undesired frequencies.
  tx_ssb_step5();
// ************* Step 6. *************
// Go back to the time domain and store the signal in alctimf.
// Remember that we use sin squared windows and that
// the transforms overlap with 50%.
// Compute power and store the peak power with
// an exponential decay corresponding to a time constant
// of 50 milliseconds.
// We expand the fft size from mic_fftsize to tx_fftsiz2 because
// the fractional resampling that will follow this step needs
// the signal to be oversampled by a factor of four to avoid
// attenuation at high frequencies. The polynomial fitting
// works as a low pass filter and we do not want any attenuation
// at the high side of our passband.
  while(clipfft_px != clipfft_pa)
    {
    pmax=tx_ssb_step6(&prat);
    lir_sched_yield();
    prat/=alc_sizhalf;
    old_prat2[old_display_ptr]=prat;
    old_pmax2[old_display_ptr]=pmax;
    prat=0;
    for(i=0; i<DISPLAY_HOLD_SIZE; i++)
      {
      prat+=old_prat2[i];
      }
    prat/=DISPLAY_HOLD_SIZE;  
    for(i=0; i<DISPLAY_HOLD_SIZE; i++)
      {
      if(pmax<old_pmax2[i])pmax=old_pmax2[i];
      }
    if(prat>0.000000001 && pmax > prat)
      {
      prat=10*log10(pmax/prat);  
      }
    else
      {
      prat=99;
      } 
    sprintf(s,"[%.1f]  ",prat);
    lir_text(28+TX_BAR_SIZE/text_width, TX_BAR_RF1CLIP_LINE, s);
    if(pmax > 1)
      {  
      pmax=2*log10(pmax);
      if(pmax>1)pmax=1;
      alc_pix2=pmax*TX_BAR_SIZE;
      }
    else
      {
      alc_pix2=0;
      }
    tx_bar(16,TX_BAR_ALC_LINE,alc_pix1,alc_pix2);
    alc_pix1=alc_pix2;
    }
// ************* Step 7. *************
// Use the slowly decaying bi-directional peak power as an ALC
// but make sure we allow at least two data blocks un-processed
// to allow the backwards fall-off to become re-calculated.
// Using the slowly varying function for an ALC (= AM modulation)
// will not increase the bandwidth by more than the bandwidth
// of the modulation signal (50 ms or 20 Hz)
  while( ((alctimf_pa-alctimf_pb+alc_bufsiz)&alc_mask) > 3*alc_fftsize)
    {
    pmax=tx_ssb_step7(&prat);
    prat/=alc_sizhalf;
    old_prat3[old_display_ptr]=prat;
    old_pmax3[old_display_ptr]=pmax;
    prat=0;
    for(i=0; i<DISPLAY_HOLD_SIZE; i++)
      {
      prat+=old_prat3[i];
      }
    prat/=DISPLAY_HOLD_SIZE;  
    for(i=0; i<DISPLAY_HOLD_SIZE; i++)
      {
      if(pmax<old_pmax3[i])pmax=old_pmax3[i];
      }
    if(prat>0.000000001 && pmax > prat)
      {
      prat=10*log10(pmax/prat);  
      }
    else
      {
      prat=99;
      } 
    sprintf(s,"[%.1f]  ",prat);
    lir_text(28+TX_BAR_SIZE/text_width, TX_BAR_ALC_LINE, s);
    }
// ************* Step 8. *************
// In case we have enough signal in the buffer, start the output.
  tx_ssb_step8();
  test_keyboard();
  if(lir_inkey != 0)
    {
    process_current_lir_inkey();
    if(lir_inkey=='X')goto end_tx_setup;
    if(lir_inkey==ARROW_UP_KEY)
      {
      txtrace_gain*=1.33;
      goto continue_setup;
      }
    if(lir_inkey==ARROW_DOWN_KEY)
      {
      txtrace_gain/=1.25;
      goto continue_setup;
      }
      
    close_tx_output();
    linrad_thread_stop_and_join(THREAD_TX_INPUT);
    close_tx_input();
    free(txmem_handle);
    lir_mutex_destroy();
    if(lir_inkey=='S')
      {
      clear_screen();
      save_tx_parms(TRUE);
      }
    if(lir_inkey=='R')
      {
      clear_screen();
      read_txpar_file();
      }
    switch (setup_mode)
      {
      case TX_SETUP_AD:
      if(lir_inkey=='L')
        {
        lir_text(0,TXPAR_INPUT_LINE,"Low:");
        txssb.minfreq=lir_get_integer(5,TXPAR_INPUT_LINE,4,1,SSB_MINFQ_HIGH);
        }
      if(lir_inkey=='H')
        {
        lir_text(0,TXPAR_INPUT_LINE,"High:");
        txssb.maxfreq=lir_get_integer(6,TXPAR_INPUT_LINE,5,
                                            SSB_MAXFQ_LOW,ui.tx_ad_speed/2);
        }
      if(lir_inkey=='F')
        {
        lir_text(0,TXPAR_INPUT_LINE,"Slope:");
        txssb.slope=lir_get_integer(7,TXPAR_INPUT_LINE,4,SSB_MINSLOPE,
                                                               SSB_MAXSLOPE);
        }
      if(lir_inkey=='B')
        {
        lir_text(0,TXPAR_INPUT_LINE,"Bass:");
        txssb.bass=lir_get_integer(6,
                                 TXPAR_INPUT_LINE,4,SSB_MINBASS,SSB_MAXBASS);
        }
      if(lir_inkey=='T')
        {
        lir_text(0,TXPAR_INPUT_LINE,"Treble:");
        txssb.treble=lir_get_integer(8,
                             TXPAR_INPUT_LINE,4,SSB_MINTREBLE,SSB_MAXTREBLE);
        }
      break;

      case TX_SETUP_MUTE:
      if(lir_inkey=='F')
        {
        lir_text(0,TXPAR_INPUT_LINE,"Freq domain:");
        txssb.mic_f_threshold=lir_get_integer(13,
                                          TXPAR_INPUT_LINE,3,0,SSB_MAX_MICF);
        }
      if(lir_inkey=='T')
        {
        lir_text(0,TXPAR_INPUT_LINE,"Time domain:");
        txssb.mic_t_threshold=lir_get_integer(13,
                                          TXPAR_INPUT_LINE,3,0,SSB_MAX_MICT);
        }
      break;

      case TX_SETUP_MICAGC:
      if(lir_inkey=='V')
        {
        lir_text(0,TXPAR_INPUT_LINE,"Volume:");
        txssb.mic_gain=lir_get_integer(8,
                                      TXPAR_INPUT_LINE,3,0,SSB_MAX_MICGAIN);
        }
      if(lir_inkey=='T')
        {
        lir_text(0,TXPAR_INPUT_LINE,"Time constant:");
        t1=lir_get_float(15,TXPAR_INPUT_LINE,6,0,SSB_MAX_MICAGC_TIME);
        txssb.mic_agc_time=100*t1;  
        }
      break;



      case TX_SETUP_RF1AGC:
      if(lir_inkey=='B')
        {
        lir_text(0,TXPAR_INPUT_LINE,"Gain:");
        txssb.mic_out_gain=lir_get_integer(15,
                                   TXPAR_INPUT_LINE,2,0,SSB_MAX_MICOUT_GAIN);
        }
      break;

      case TX_SETUP_RF1CLIP:
      if(lir_inkey=='B')
        {
        lir_text(0,TXPAR_INPUT_LINE,"Clipper gain:");
        txssb.rf1_gain=lir_get_integer(15,
                       TXPAR_INPUT_LINE,3,SSB_MIN_RF1_GAIN,SSB_MAX_RF1_GAIN);
        }
      break;
      }
    if(lir_inkey=='A')setup_mode=TX_SETUP_MICAGC;
    if(lir_inkey=='M')setup_mode=TX_SETUP_AD;
    if(lir_inkey=='Q')setup_mode=TX_SETUP_MUTE;
    if(lir_inkey=='D')setup_mode=TX_SETUP_RF1AGC;
    if(lir_inkey=='C')setup_mode=TX_SETUP_RF1CLIP;
    if(lir_inkey=='+')
      {
      tg.spproc_no++;
      if(tg.spproc_no > MAX_SSBPROC_FILES)tg.spproc_no=MAX_SSBPROC_FILES;
      }
    if(lir_inkey=='-')
      {
      tg.spproc_no--;
      if(tg.spproc_no < 0)tg.spproc_no=0;
      }
    goto restart;
    }
continue_setup:;  
  }
end_tx_setup:;
close_tx_output();
linrad_thread_stop_and_join(THREAD_TX_INPUT);
close_tx_input();
free(txmem_handle);
lir_mutex_destroy();
tg.spproc_no=default_spproc_no;
}

void use_tx_resamp(float ampfac)
{
int i, j, k, m;
float t1, t2, t3, r1, r2;
// The latest half block of data resides in tx_resamp[alc_fftsize]
// to tx_resamp[2*alc_fftsize-1]
// Copy the previous block from resamp_tmp and multiply with
// the window function.
k=alc_sizhalf;
m=k-1;
t1=0;
for(i=0; i<alc_sizhalf; i++)
  {
  tx_resamp[2*i  ]=resamp_tmp[2*i  ]*alc_win[i]; 
  tx_resamp[2*i+1]=resamp_tmp[2*i+1]*alc_win[i]; 
  resamp_tmp[2*i  ]=ampfac*tx_resamp[2*k  ];
  resamp_tmp[2*i+1]=ampfac*tx_resamp[2*k+1];
  tx_resamp[2*k  ]*=ampfac*alc_win[m];
  tx_resamp[2*k+1]*=ampfac*alc_win[m];
  m--;
  k++;
  } 
fftforward(alc_fftsize, alc_fftn, tx_resamp, alc_table, alc_permute);
// Clear the spectrum outside the desired passband.
// These data points should be very small.
k=txout_fftsize-alc_fftsize;
if(k > 0)
  {
  for(i=alc_sizhalf; i<alc_fftsize; i++)
    {
    tx_resamp[2*(k+i)  ]=tx_resamp[2*i  ];
    tx_resamp[2*(k+i)+1]=tx_resamp[2*i+1];
    }
  k=txout_fftsize-alc_sizhalf;
  for(i=alc_sizhalf; i<k; i++)
    {
    tx_resamp[2*i  ]=0;
    tx_resamp[2*i+1]=0;
    }
  i=tx_filter_ib3+1;
  k=tx_filter_ia3-1;
  t1=1;
  t3=4./tx_filter_ib3;
  t2=0;
  while(k>=i && t1 >0)
    {
    tx_resamp[2*k]*=t1;
    tx_resamp[2*k+1]*=t1;
    tx_resamp[2*i]*=t1;
    tx_resamp[2*i+1]*=t1;
    t2+=t3;
    t1=1-t2*t2;
    i++;
    k--;
    }
  while(k>=i)
    {
    tx_resamp[2*k]=0;
    tx_resamp[2*k+1]=0;
    tx_resamp[2*i]=0;
    tx_resamp[2*i+1]=0;
    i++;
    k--;
    }
  }
if(k < 0)
  {
  lirerr(77676);
  }
//
//if(screen_count>=MAX_SCREENCOUNT)
//  {
//  if(tx_setup_flag == TRUE)show_txfft(tx_resamp,0,2,txout_fftsize);
//  }
// Now we have the output in the frequency domain at the size txout_fftsize
// with a sampling rate that fits our D/A converter. Transform it back
// to the time domain into txout.
fftback(txout_fftsize, txout_fftn, tx_resamp, txout_table, txout_permute);
for(i=0; i<txout_fftsize; i+=2)
  {
  tx_output_phase+=tx_output_phstep;
  if(tx_output_phase > PI_L)tx_output_phase-=2*PI_L;
  if(tx_output_phase < -PI_L)tx_output_phase+=2*PI_L;
  t1=cos(tx_output_phase);
  t2=sin(tx_output_phase);
  r1=tx_resamp[i  ]+txout_tmp[i  ];
  r2=tx_resamp[i+1]+txout_tmp[i+1];
  txout[i  ]=r1*t1+r2*t2;
  txout[i+1]=-r1*t2+r2*t1;
  }
for(i=0; i<txout_fftsize; i++)
  {
  txout_tmp[i]=tx_resamp[txout_fftsize+i];
  }
// The microphone input is real data at the soundcard speed.
// For timing we use the complex samples at half the speed.
tx_wsamps=lir_tx_input_samples()/2;
k=(mictimf_pa-mictimf_px+mictimf_bufsiz)&mictimf_mask;
k+=(micfft_pa-micfft_px+micfft_bufsiz)&micfft_mask;
k+=(cliptimf_pa-cliptimf_px+micfft_bufsiz)&micfft_mask; 
j=(clipfft_pa-clipfft_px+alc_bufsiz)&alc_mask;
j+=(alctimf_pa-((int)(alctimf_fx))+4+alc_bufsiz)&alc_mask;
j+=alc_fftsize;
k+=j/tx_pre_resample;
tx_wsamps+=k/(2*sizeof(float));
// The maximum amplitude should ideally be 1.0 but since
// a litte power outside the passband was removed in the last 
// backwards transform, a small amount of re-peaking could occur
// Set the maximum level to TX_DA_MARGIN of full scale to make sure
// D/A converters will not overflow.

// ****************************************************
/* This is to be moved to CW hand keying.....
t1=(float)(lir_tx_output_samples())/
                     (tx_output_upsample*tx_resample_ratio*ui.tx_ad_speed);
r1=(float)(txout_sizhalf)/
                     (tx_output_upsample*tx_resample_ratio*ui.tx_ad_speed);

hware_hand_key();
if(old_hand_key != hand_key)
  {
  hware_set_rxtx(hand_key);
  old_hand_key=hand_key;
  }

if(old_hand_key == FALSE)
  {
  for(i=0; i<txout_fftsize; i++)
    {
    txout[i  ]=0; 
    }
  }

t1=0;
for(i=0; i<txout_sizhalf; i++)
  {
  t2=txout[2*i  ]*txout[2*i  ]+
     txout[2*i+1]*txout[2*i+1];
  if(t2>t1)t1=t2;
fprintf(dmp,"\n%d   %f",i,sqrt(t2));
  }     
*/

switch(tx_output_mode)
  {
  case 0:
  for(i=0; i<txout_sizhalf; i++)
    {
    t1=txout[2*i  ];
    tx_out_shi[i]=TX_DA_MARGIN*0x7fff*t1;
    }
  write(tx_audio_out,tx_out_shi,tx_write_bytes);
  break;
  
  case 1:
  for(i=0; i<txout_sizhalf; i++)
    {
    t1=txout[2*i  ];
    t2=txout[2*i+1];
    tx_out_shi[2*i]=TX_DA_MARGIN*0x7fff*t1;
    tx_out_shi[2*i+1]=TX_DA_MARGIN*0x7fff*t2;
    }
  write(tx_audio_out,tx_out_shi,tx_write_bytes);
  break;

  case 2:
  for(i=0; i<txout_sizhalf; i++)
    {
    t1=txout[2*i  ];
    tx_out_int[i]=TX_DA_MARGIN*0x7fff0000*t1;
    }
  write(tx_audio_out,tx_out_int,tx_write_bytes);
  break;
  
  case 3:
  for(i=0; i<txout_sizhalf; i++)
    {
    t1=txout[2*i  ];
    t2=txout[2*i+1];
    tx_out_int[2*i  ]=TX_DA_MARGIN*0x7fff0000*t1;
    tx_out_int[2*i+1]=TX_DA_MARGIN*0x7fff0000*t2;
    }
  write(tx_audio_out,tx_out_int,tx_write_bytes);
  break;
  }
tx_wsamps+=(float)(lir_tx_output_samples())/
                      (tx_output_upsample*tx_resample_ratio);
tx_wttim=(float)(tx_wsamps)/ui.tx_ad_speed;



tx_wttim_sum+=tx_wttim;
tx_wrblock_no++;
if(tx_wrblock_no == TX_WRBLOCK_MAXCNT)
  {
  tx_wrblock_no=0;
  tx_wttim_sum/=TX_WRBLOCK_MAXCNT;
  if(tx_wttim_ref == 0)
    {
    tx_wttim_ref=tx_wttim_sum;
    tx_wttim_diff=0;
    }
  else
    {
    t1=tx_wttim_sum-tx_wttim_ref;
    t2=t1-tx_wttim_diff;
    tx_wttim_diff=t1;
// We have accumulated an error of t1 seconds while writing 
// TX_WRBLOCK_MAXCNT data blocks. Each block spans a time of
// txout_sizhalf/txssb.output speed so the relative error is
// X=t1/(TX_WRBLOCK_MAXCNT*txout_sizhalf/txssb.output speed);
// Therefore we should multiply the output 
// speed by (1-X)
// Use half the correction and add in a small part of t1,
// the deviation from the initial time delay.
    new_tx_resample_ratio=tx_resample_ratio*(1-0.5*(t2+0.2*t1)*
                           ui.tx_da_speed/(TX_WRBLOCK_MAXCNT*txout_sizhalf));    
    }
  tx_wttim_sum=0;
  }
lir_refresh_screen();
}


void resample_tx_output(void)
{
int i1, i2, i3, i4, mask2;
float t1, t2, t3, t4, t5, t6, t7, r1, r2;
float a1, a2, rdiff;
// The microphone is best sampled at a low sampling speed since
// we do not want much bandwidth nor dynamic range.
// For the output the requirements depend on the circumstances.
// In case the transmitter is used in a wideband up-conversion
// system, the output sampling rate may be very high and the
// requirements on the signal purity might be very high.
// We do the resampling in two steps. First do a fractional
// resampling to a speed near the speed of the microphone
// signal (but perhaps a power of two below the desired output
// speed as indicated by tx_output_upsample.)
// Subsequently, use an FFT to sample up the signal to the
// desired frequency while filtering out the false signals
// that non-linearities in the fractional resampling have
// introduced.
mask2=alc_mask-1;
begin:;
r1=tx_resamp_pa/tx_resample_ratio;
resamp:;
r2=(tx_resamp_pa+2)/tx_resample_ratio;
i2=alctimf_fx+r1+1;
i3=alctimf_fx+r2+1;
i2&=mask2;
i3&=mask2;
if(abs(i3-i2) > 2)
  {
  i2=alctimf_fx+(r1+r2)/2;
  i3=i2+2;
  i2&=mask2;
  i3&=mask2;
  }
else
  {
  if(i3==i2)
    {
    i2=alctimf_fx+r1;
    i2&=mask2;
    if(i3==i2)
      {
      i3=i2+2;
      i3&=mask2;
      }
    }
  }        
i4=(i3+2)&(mask2);
i1=(i2+mask2)&(mask2);
if( ((alctimf_pb-i4+alc_bufsiz)&alc_mask) > 4)
  {
  if(alctimf_mute[i2/alc_fftsize]==FALSE ||
     alctimf_mute[i3/alc_fftsize]==FALSE)
    {
    rdiff=r1+alctimf_fx-i2;
    if(rdiff > alc_bufsiz/2)
      {
      rdiff-=alc_bufsiz;
      }
    rdiff/=2; 
// Use Lagrange's interpolation formula to fit a third degree
// polynomial to 4 points:
//  a1=-rdiff *   (rdiff-1)*(rdiff-2)*alctimf[i1]/6
//     +(rdiff+1)*(rdiff-1)*(rdiff-2)*alctimf[i2]/2
//     -(rdiff+1)* rdiff   *(rdiff-2)*alctimf[i3]/2
//     +(rdiff+1)* rdiff   *(rdiff-1)*alctimf[i4]/6; 
// Rewrite slightly to save a few multiplications - do not
// think the compiler is smart enough to do it for us.
    t1=rdiff-1;
    t2=rdiff-2;
    t3=rdiff+1;
    t4=t1*t2;
    t5=t3*rdiff;
    t6=rdiff*t4;
    t4=t3*t4;
    t7=t5*t2;
    t5=t5*t1;
    a1=((t5*alctimf[i4  ]-t6*alctimf[i1  ])/3+t4*alctimf[i2  ]-t7*alctimf[i3  ])/2;
    a2=((t5*alctimf[i4+1]-t6*alctimf[i1+1])/3+t4*alctimf[i2+1]-t7*alctimf[i3+1])/2;
// The curve fitting is (of course) just an approximation.
// Make sure we do not go outside our range!
    t2=a1*a1+a2*a2;
    if(t2 > tx_resamp_ampfac1)tx_resamp_ampfac1=t2;
    tx_resamp[tx_resamp_pa+alc_fftsize  ]=a1;
    tx_resamp[tx_resamp_pa+alc_fftsize+1]=a2;
    }
  else
    {
    tx_resamp[tx_resamp_pa+alc_fftsize  ]=0;
    tx_resamp[tx_resamp_pa+alc_fftsize+1]=0;
    }
  r1=r2;
  tx_resamp_pa+=2;
  if(tx_resamp_pa >= alc_fftsize)
    {
    alctimf_fx+=r2;
    if(alctimf_fx>alc_bufsiz)alctimf_fx-=alc_bufsiz;
    t1=tx_resamp_ampfac1;
    if(t1 < tx_resamp_ampfac2)t1=tx_resamp_ampfac2;
/*
{int i;
for(i=0; i<alc_sizhalf; i++)
  {
  t2=tx_resamp[alc_fftsize+2*i  ]*tx_resamp[alc_fftsize+2*i  ]+
     tx_resamp[alc_fftsize+2*i+1]*tx_resamp[alc_fftsize+2*i+1];
fprintf(dmp,"\n%d  %f ",i,sqrt(t2));
  }   }  
*/

    use_tx_resamp(sqrt(1/t1));
    tx_resamp_ampfac2=tx_resamp_ampfac1;
    tx_resamp_ampfac1=1;
    tx_resample_ratio=new_tx_resample_ratio;
    tx_resamp_pa=0;
    goto begin;
    }
  goto resamp;  
  }
}



void tx_bar(int xt,int yt,int val1, int val2)
{
int x, y;
x=xt*text_width;
y=yt*text_height+1;
if(val1 == val2)
  {
  if(val1 >= 0)return;
  lir_fillbox(x,y,TX_BAR_SIZE,text_height-2,7);
  return;
  }
if(val2 > val1)
  { 
  lir_fillbox(x+val1,y,val2-val1,text_height-2,12);
  }
else
  {  
  lir_fillbox(x+val2,y,val1-val2,text_height-2,7);
  } 
}


void show_txfft(float *z, float lim, int type, int siz)
{
int i, ia, ib, k, m;
int pixels_per_bin, bins_per_pixel;
float t1,t2;
char color;
short int *trc;
if(type >= MAX_DISPLAY_TYPES)
  {
  lirerr(762319); 
  return;
  }
trc=&txtrace[type*tx_show_siz];  
ia=0;
t2=0;
pixels_per_bin=(screen_width-1)/siz;
bins_per_pixel=1;
while(pixels_per_bin == 0)
  {
  siz/=2;
  pixels_per_bin=(screen_width-1)/siz;
  bins_per_pixel*=2;
  }
ib=pixels_per_bin;
lir_sched_yield();
for(i=1; i<siz; i++)
  {
  lir_line(ia,trc[i-1],ib,trc[i],0);
  ia=ib;
  ib+=pixels_per_bin;
  }
lir_sched_yield();
ia=0;
ib=pixels_per_bin;
m=0;
mailbox[0]=0;
for(i=0; i<siz; i++)
  {
  t1=0;
  for(k=0; k<bins_per_pixel; k++)
    {
    t1=z[2*m]*z[2*m]+z[2*m+1]*z[2*m+1];
    m++;
    }
  t1/=bins_per_pixel;  
  k=-0.07*log10(txtrace_gain*(t1+0.0000000000001))*screen_height;
  if(k>txtrace_height)k=txtrace_height;
  if(k<0)k=0;
  trc[i]=k+screen_height-txtrace_height-1;
  if(i>0)
    {
    if(type==0)
      {
      if(t2 > lim)
        {
        color=15;
        }
      else
        {
        color=12;
        }
      }
    else
      {
      color=display_color[type];
      }      
    lir_line(ia,trc[i-1],ib,trc[i],color);
    ia=ib;
    ib+=pixels_per_bin;
    }
  t2=t1;
  }
lir_sched_yield();
}

void save_tx_parms(char verbose)
{
int *wgi;
char s[80];
char **partext;
int i, nn, line;
FILE *txpar_file;
wgi=NULL;
nn=0;
partext=NULL;
line=4;
make_txproc_filename();
txpar_file=fopen(txproc_filename,"w");
if(txpar_file == NULL)
  {
  lirerr(381164);
  return;
  }
switch (rx_mode)
  {
  case MODE_SSB:
  wgi=(void*)(&txssb);
  txssb.check=SSBPROC_VERNR;
  partext=ssbprocparm_text;
  nn=MAX_SSBPROC_PARM;
  break;

  case MODE_WCW:
  case MODE_NCW:
  case MODE_HSMS:
  case MODE_QRSS:
  wgi=(void*)(&txcw);
  partext=cwprocparm_text;
  nn=MAX_CWPROC_PARM;
  break;
  
  default:
  lirerr(918453);
  break;
  }
for(i=0; i<nn; i++)
  {
  sprintf(s,"%s [%d]",partext[i],wgi[i]);
  fprintf(txpar_file,"%s\n",s);
  if(verbose && line<screen_last_line)lir_text(5,line,s);
  line++;
  }
parfile_end(txpar_file);
sprintf(s,"Parameters saved to %s.   %s",txproc_filename, press_any_key);
lir_text(10,line+3,s);  
await_keyboard();
}



int check_tx_devices(void)
{
if(ui.tx_addev_no < 0 ||
   ui.tx_addev_no >= MAX_DEVNAMES ||
   ui.tx_dadev_no < 0 ||
   ui.tx_dadev_no >= MAX_DEVNAMES ||
   ui.tx_da_channels < 1 ||
   ui.tx_da_channels > 2 ||
   (ui.tx_da_bytes !=2 && ui.tx_da_bytes !=4) ||
   (ui.tx_ad_bytes !=2 && ui.tx_ad_bytes !=4) ) return FALSE;
return TRUE;
}


int read_txpar_file(void)
{
int i, j, k, nn;
int *wgi;
int rdbytes;
FILE *txpar_file;
char *testbuff;
char **partext;
partext=NULL;
wgi=NULL;
nn=0;
if(check_tx_devices() == FALSE)return FALSE;
testbuff = malloc(4096);
if(testbuff == NULL)
  {
  lirerr(1286);
  return FALSE;
  }
for(i=0; i<4096; i++)testbuff[i]=0;  
make_txproc_filename();
if(txproc_filename == NULL)return FALSE;
txpar_file=fopen(txproc_filename,"r");
if(txpar_file != NULL)
  {
  rdbytes=fread(testbuff,1,4095,txpar_file);
  fclose(txpar_file);
  if(rdbytes >= 4095)goto reterr;
  switch (rx_mode)
    {
    case MODE_SSB:
    wgi=(void*)(&txssb);
    partext=ssbprocparm_text;
    nn=MAX_SSBPROC_PARM;
    break;

    case MODE_WCW:
    case MODE_NCW:
    case MODE_HSMS:
    case MODE_QRSS:
    wgi=(void*)(&txcw);
    partext=cwprocparm_text;
    nn=MAX_CWPROC_PARM;
    break;
  
    default:
    lirerr(918453);
    break;
    }

  k=0;
  for(i=0; i<nn; i++)
    {
    while( (testbuff[k]==' ' || testbuff[k]== '\n' ) && k<rdbytes)k++;
    j=0;
    while(testbuff[k]== partext[i][j] && k<rdbytes)
      {
      k++;
      j++;
      } 
    if(partext[i][j] != 0)goto reterr;    
    while(testbuff[k]!='[' && k<rdbytes)k++;
    if(k>=rdbytes)goto reterr;
    sscanf(&testbuff[k],"[%d]",&wgi[i]);
    while(testbuff[k]!='\n' && k<rdbytes)k++;
    if(k>=rdbytes)goto reterr;
    }
  switch (rx_mode)
    {
    case MODE_SSB:
    if(txssb.check!=SSBPROC_VERNR ||
       txssb.minfreq < SSB_MINFQ_LOW || 
       txssb.minfreq > SSB_MINFQ_HIGH ||
       txssb.maxfreq < SSB_MAXFQ_LOW || 
       txssb.maxfreq > SSB_MAXFQ_HIGH ||
       txssb.slope > SSB_MAXSLOPE || 
       txssb.slope < SSB_MINSLOPE ||
       txssb.bass > SSB_MAXBASS || 
       txssb.bass < SSB_MINBASS ||
       txssb.treble > SSB_MAXTREBLE || 
       txssb.treble < SSB_MINTREBLE ||
       txssb.mic_f_threshold < 0 ||
       txssb.mic_f_threshold > SSB_MAX_MICF  ||
       txssb.mic_t_threshold < 0 ||
       txssb.mic_t_threshold > SSB_MAX_MICT ||
       txssb.mic_gain < 0 ||
       txssb.mic_gain > SSB_MAX_MICGAIN ||
       txssb.mic_agc_time < 0 ||
       txssb.mic_agc_time > SSB_MAX_MICAGC_TIME ||
       txssb.mic_out_gain < 0 ||
       txssb.mic_out_gain > SSB_MAX_MICOUT_GAIN ||
       txssb.rf1_gain < SSB_MIN_RF1_GAIN ||
       txssb.rf1_gain > SSB_MAX_RF1_GAIN )
      {
      goto reterr;
      }
    break;

    case MODE_WCW:
    case MODE_NCW:
    case MODE_HSMS:
    case MODE_QRSS:
    break;
    }
  }  
else
  {
reterr:;
  free(testbuff);  
  return FALSE;
  }  
free(testbuff);  
return TRUE;
}

void set_default_spproc_parms(void)
{
txssb.minfreq=200;
txssb.maxfreq=2600;
txssb.slope=6;
txssb.bass=-8;
txssb.treble=0;
txssb.mic_f_threshold=15;
}

void check_tg_borders(void)
{
current_graph_minh=TG_VSIZ;
current_graph_minw=TG_HSIZ;
check_graph_placement((void*)(&tg));
set_graph_minwidth((void*)(&tg));
}

void new_ssbproc_no(void)
{
int oldno;
oldno=tg.spproc_no;
tg.spproc_no=numinput_int_data;
if(read_txpar_file() == FALSE)
  {
  tg.spproc_no=oldno;
  read_txpar_file();
  }
sc[SC_SHOW_TX_FQ]++;
make_modepar_file(GRAPHTYPE_TG);
}   

void new_tx_frequency(void)
{
float t1;
tg.freq=numinput_float_data;
t1=1000000*tg.freq;
if(t1<-46000)t1=-46000;
if(t1>46000)t1=46000;
tx_output_phstep=2*PI_L*(t1+tx_ssbproc_fqshift)/ui.tx_da_speed;
sc[SC_SHOW_TX_FQ]++;
make_modepar_file(GRAPHTYPE_TG);
}


void help_on_tx_graph(void)
{
int msg_no;
int event_no;
// Set msg invalid in case we are not in any select area.
msg_no=-1;
for(event_no=0; event_no<MAX_TGBUTT; event_no++)
  {
  if( tgbutt[event_no].x1 <= mouse_x && 
      tgbutt[event_no].x2 >= mouse_x &&      
      tgbutt[event_no].y1 <= mouse_y && 
      tgbutt[event_no].y2 >= mouse_y) 
    {
    switch (event_no)
      {
      case TG_TOP:
      case TG_BOTTOM:
      case TG_LEFT:
      case TG_RIGHT:
      msg_no=101;
      break;

      case TG_INCREASE_FQ: 
      msg_no=104;
      break;
  
      case TG_DECREASE_FQ:
      msg_no=103;
      break;

      case TG_NEW_TX_FREQUENCY:
      msg_no=105;
      break;

      case TG_CHANGE_SSBPROC_NO:
      msg_no=106;
      break;
      }
    }  
  }
help_message(msg_no);
}




void mouse_continue_tx_graph(void)
{
int i, j;
switch (mouse_active_flag-1)
  {
  case TG_TOP:
  if(tg.ytop!=mouse_y)goto tgm;
  break;

  case TG_BOTTOM:
  if(tg.ybottom!=mouse_y)goto tgm;
  break;

  case TG_LEFT:
  if(tg.xleft!=mouse_x)goto tgm;
  break;

  case TG_RIGHT:
  if(tg.xright==mouse_x)break;
tgm:;
  pause_screen_and_hide_mouse();
  dual_graph_borders((void*)&tg,0);
  if(tg_oldx==-10000)
    {
    tg_oldx=mouse_x;
    tg_oldy=mouse_y;
    }
  else
    {
    i=mouse_x-tg_oldx;
    j=mouse_y-tg_oldy;  
    tg_oldx=mouse_x;
    tg_oldy=mouse_y;
    tg.ytop+=j;
    tg.ybottom+=j;      
    tg.xleft+=i;
    tg.xright+=i;
    check_tg_borders();
    }
  graph_borders((void*)&tg,15);
  resume_thread(THREAD_SCREEN);
  break;
    
  default:
  goto await_release;
  }
if(leftpressed == BUTTON_RELEASED)goto finish;
return;
await_release:;
if(leftpressed != BUTTON_RELEASED) return;
if(ui.network_flag != 2)
  {
  switch (mouse_active_flag-1)
    {
    case TG_INCREASE_FQ: 
    if(tx_hware_init_flag == 0)
      {
      tg.freq+=fg.passband_increment;
      }
    break;
  
    case TG_DECREASE_FQ:
    if(tx_hware_init_flag == 0)
      {
      tg.freq-=fg.passband_increment;
      }
    break;

    case TG_NEW_TX_FREQUENCY:
    if(tx_hware_init_flag == 0)
      {
      numinput_xpix=tgbutt[TG_NEW_TX_FREQUENCY].x1+3;
      numinput_ypix=tgbutt[TG_NEW_TX_FREQUENCY].y2-text_height+1;
      numinput_chars=FREQ_MHZ_DIGITS+7;
      erase_numinput_txt();
      numinput_flag=FIXED_FLOAT_PARM;
      par_from_keyboard_routine=new_tx_frequency;
      }
    return;  
    break;
    
    case TG_CHANGE_SSBPROC_NO:
    if(tx_hware_init_flag == 0)
      {
      numinput_xpix=tgbutt[TG_CHANGE_SSBPROC_NO].x1+3+5*text_width;
      numinput_ypix=tgbutt[TG_CHANGE_SSBPROC_NO].y2-text_height+1;
      numinput_chars=2;
      erase_numinput_txt();
      numinput_flag=FIXED_INT_PARM;
      par_from_keyboard_routine=new_ssbproc_no;
      }
    return;  
    break;
    
    default:
    lirerr(872);
    break;
    }
  }  
finish:;
leftpressed=BUTTON_IDLE;  
mouse_active_flag=0;
make_tx_graph(TRUE);
tg_oldx=-10000;
}

void mouse_on_tx_graph(void)
{
int event_no;
// First find out is we are on a button or border line.
for(event_no=0; event_no<MAX_TGBUTT; event_no++)
  {
  if( tgbutt[event_no].x1 <= mouse_x && 
      tgbutt[event_no].x2 >= mouse_x &&      
      tgbutt[event_no].y1 <= mouse_y && 
      tgbutt[event_no].y2 >= mouse_y) 
    {
    tg_old_y1=tg.ytop;
    tg_old_y2=tg.ybottom;
    tg_old_x1=tg.xleft;
    tg_old_x2=tg.xright;
    mouse_active_flag=1+event_no;
    current_mouse_activity=mouse_continue_tx_graph;
    return;
    }
  }
// Not button or border.
current_mouse_activity=mouse_nothing;
mouse_active_flag=1;
}


void make_tx_graph(int clear_old)
{
pause_thread(THREAD_SCREEN);
if(clear_old)
  {
  hide_mouse(tg_old_x1,tg_old_x2,tg_old_y1,tg_old_y2);
  lir_fillbox(tg_old_x1,tg_old_y1,tg_old_x2-tg_old_x1+1,
                                                    tg_old_y2-tg_old_y1+1,0);
  }
tg_flag=1;
check_tg_borders();
hide_mouse(tg.xleft,tg.xright,tg.ytop,tg.ybottom);  
scro[tx_graph_scro].no=TRANSMIT_GRAPH;
scro[tx_graph_scro].x1=tg.xleft;
scro[tx_graph_scro].x2=tg.xright;
scro[tx_graph_scro].y1=tg.ytop;
scro[tx_graph_scro].y2=tg.ybottom;
tgbutt[TG_LEFT].x1=tg.xleft;
tgbutt[TG_LEFT].x2=tg.xleft;
tgbutt[TG_LEFT].y1=tg.ytop;
tgbutt[TG_LEFT].y2=tg.ybottom;
tgbutt[TG_RIGHT].x1=tg.xright;
tgbutt[TG_RIGHT].x2=tg.xright;
tgbutt[TG_RIGHT].y1=tg.ytop;
tgbutt[TG_RIGHT].y2=tg.ybottom;
tgbutt[TG_TOP].x1=tg.xleft;
tgbutt[TG_TOP].x2=tg.xright;
tgbutt[TG_TOP].y1=tg.ytop;
tgbutt[TG_TOP].y2=tg.ytop;
tgbutt[TG_BOTTOM].x1=tg.xleft;
tgbutt[TG_BOTTOM].x2=tg.xright;
tgbutt[TG_BOTTOM].y1=tg.ybottom;
tgbutt[TG_BOTTOM].y2=tg.ybottom;
// Draw the border lines
graph_borders((void*)&tg,7);
tg_oldx=-10000;
settextcolor(7);
make_button(tg.xleft+text_width,tg.ybottom-text_height/2-2,
                                       tgbutt,TG_DECREASE_FQ,25);
tgbutt[TG_NEW_TX_FREQUENCY].x1=tgbutt[TG_DECREASE_FQ].x2+0.5*text_width-2;
tgbutt[TG_NEW_TX_FREQUENCY].x2=tgbutt[TG_NEW_TX_FREQUENCY].x1+
                                      (FREQ_MHZ_DIGITS+9)*text_width;
tgbutt[TG_NEW_TX_FREQUENCY].y1=tg.ybottom-text_height-3;
tgbutt[TG_NEW_TX_FREQUENCY].y2=tgbutt[TG_NEW_TX_FREQUENCY].y1+text_height+1;
make_button(tgbutt[TG_NEW_TX_FREQUENCY].x2+text_width,
                       tg.ybottom-text_height/2-2, tgbutt,TG_INCREASE_FQ,24); 
tgbutt[TG_CHANGE_SSBPROC_NO].x1=tg.xleft+2;
tgbutt[TG_CHANGE_SSBPROC_NO].x2=tg.xleft+2+7.5*text_width;
tgbutt[TG_CHANGE_SSBPROC_NO].y1=tg.ybottom-2.5*(text_height+2);
tgbutt[TG_CHANGE_SSBPROC_NO].y2=tgbutt[TG_CHANGE_SSBPROC_NO].y1+text_height+1;

resume_thread(THREAD_SCREEN);
sc[SC_SHOW_TX_FQ]++;
make_modepar_file(GRAPHTYPE_TG);
}


void init_tx_graph(void)
{
if (read_modepar_file(GRAPHTYPE_TG) == 0)
  {
  tg.xright=screen_width-text_width;;
  tg.xleft=tg.xright-TG_HSIZ;
  tg.ybottom=screen_height-4*text_height;
  tg.ytop=tg.ybottom-TG_VSIZ;
  }
if(tg.spproc_no<0)tg.spproc_no=0;
if(tg.spproc_no>MAX_SSBPROC_FILES)tg.spproc_no=MAX_SSBPROC_FILES;
tx_hware_init_flag=1;
if(ui.tx_dadev_no < SPECIFIC_DEVICE_CODES)clear_hware_data();
tx_graph_scro=no_of_scro;
make_tx_graph(FALSE);
no_of_scro++;
if(no_of_scro >= MAX_SCRO)lirerr(89);
tx_hware_init_flag=0;
}
